/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package graphics.texture;

import java.nio.FloatBuffer;
import java.util.HashMap;
import java.util.Map;
import logging.Logger;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL30;

/**
 * Eine Textur ist eine Sammlung von Daten auf der Grafikkarte, die lesenden
 * wahlfreien Zugriff im Shader ermoeglichen.
 * @author nico3000
 */
public class Texture {    
    /**
     * Unterstuetzte Typen von Texturen.
     */
    public static enum Type {
        /**
         * Eindimensionale Textur.
         */
        TEXTURE_1D,
        /**
         * Zweidimensionale Textur.
         */
        TEXTURE_2D,
    }
    
    /**
     * Alle Instanzen der Klasse Texture nach ihrer Resource.
     */
    private static Map<String, Texture> INSTANCES = new HashMap<>();
    
    /**
     * Erzeugt eine neue initialisierte Textur fuer eine bestimmte Resource.
     * @param type Der Typ der Textur
     * @param resource Die Resource
     * @return Die erstellte Textur oder <tt>null</tt>, falls ein Fehler
     *         aufgetreten ist
     */
    public static Texture createTexture(Type type, String resource) {
        Texture tex = INSTANCES.get(resource);
        if(tex == null) {
            tex = new Texture(type, resource);
            if(!tex.init()) {
                Logger.INSTANCE.error("Initialization of texture from resource %s failed!", resource);
                return null;
            }
            INSTANCES.put(resource, tex);
            tex.addRef();
            return tex;
        } else {
            Logger.INSTANCE.error("Texture from %s already exists!", resource);
            return null;
        }
    }
    
    /**
     * Liefert eine Textur, sofern sie vorher mit <tt>createTexture()</tt>
     * erfolgreich erzeugt wurde.
     * @param resource Die Resource
     * @return Die Textur
     */
    public static Texture getTexture(String resource) {
        Texture tex = INSTANCES.get(resource);
        if(tex != null) {
            if(!tex.isInitialized() && !tex.init()) {
                Logger.INSTANCE.error("Reinitializing of Texture %s failed!", resource);
                return null;
            }
            tex.addRef();
        }
        return tex;
    }
    
    /**
     * Gibt den Status der bekannten Texturen zurueck.
     * @return Der Status
     */
    public static String getDebugStatus() {
        String textures = "";
        for(Texture tex : INSTANCES.values()) {
            if(tex.isInitialized()) {
                textures += String.format("\n- %s, reference count = %d", tex.getResource(), tex.ref);
                if(!tex.hasData()) {
                    textures += " (no data)";
                }
            }
        }
        return String.format("Active Texture instances: %s", textures.isEmpty() ? "none" : textures);
    }
    
    /**
     * Die Slots und die momentan daran gebundenen Texturen.
     */
    private static final Texture bindings[] = new Texture[80];
    /**
     * Der als Working Slot definierte Slot.
     */
    private static final int WORKING_SLOT = 0;
    /**
     * Der momentan aktive Slot.
     */
    private static int activeSlot = -1;
    
    /**
     * Die ID der Textur.
     */
    private int id = 0;
    /**
     * Das Binding Target der Textur.
     */
    private final int target;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * Groesse der Textur im VRAM in Byte.
     */
    private int size = 0;
    /**
     * Die Breite der Textur in Pixeln.
     */
    private int width = 0;
    /**
     * Die Hoehe der Textur in Pixeln.
     */
    private int height = 0;
    /**
     * Die Resource der Textur.
     */
    private final String resource;
    /**
     * Gibt an, ob die Textur bereits Daten hat.
     */
    private boolean hasData = false;

    /**
     * Konstruktor.
     * @param type Der Typ der Textur
     */
    private Texture(Type type, String resource) {
        switch(type) {
            case TEXTURE_1D:
                this.target = GL11.GL_TEXTURE_1D;
                break;
            case TEXTURE_2D:
                this.target = GL11.GL_TEXTURE_2D;
                break;
            default:
                // should not happen
                this.target = 0;
                break;
        }
        this.resource = resource;
    }
    
    /**
     * Liefert die Resource der Textur.
     * @return Die Resource
     */
    public String getResource() {
        return resource;
    }
    
    /**
     * Initialisiert die Textur.
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean init() {
        this.destroy();
        this.id = GL11.glGenTextures();
        return this.id != 0;
    }
    
    /**
     * Zerstoert die Textur und ihre Daten.
     */
    private void destroy() {
        if(this.id != 0) {
            //Logger.INSTANCE.info("Now destroying texture %s.", this.resource);
            GL11.glDeleteTextures(this.id);
            this.id = 0;
            this.hasData = false;
        }
    }
    
    private void addRef() {
        ++this.ref;
    }
    
    /**
     * Gibt die Textur wieder frei.
     */
    public void release() {
        if(this.ref == 0) {
            Logger.INSTANCE.warning("Releasing texture %s impossible because reference counter is already 0.", this.resource);
        } else if(--this.ref == 0) {
            this.destroy();
        }
    }

    /**
     * Liefert die OpenGL Konstante des Targets zurueck.
     * @return Das Target
     */
    public int getTarget() {
        return target;
    }
    
    /**
     * Bindet die Textur an ihr Target.
     */
    public void bind() {
        activateSlot(WORKING_SLOT);
        GL11.glBindTexture(this.target, this.id);
    }
    
    /**
     * Verbindet diese Textur mit ihrem Slot, falls sie es noch nicht ist.
     */
    public void connectToSlot(int slot) {
        if(bindings[slot] != this) {
            activateSlot(slot);
            GL11.glBindTexture(this.target, this.id);
            bindings[slot] = this;
        }
    }
    
    /**
     * Aktiviert einen Slot, falls er nicht schon aktiv ist.
     * @param slot Der zu aktivierende Slot
     */
    private static void activateSlot(int slot) {
        if(slot != activeSlot) {
            GL13.glActiveTexture(GL13.GL_TEXTURE0 + slot);
            activeSlot = slot;
        }
    }
    
    /**
     * Setzt die Daten der Textur.
     * @param width Die Breite
     * @param height Die Hoehe
     * @param channels Die Anzahl der Farbkanaele
     * @param data Die Daten
     */
    public void setData(int width, int height, int channels, FloatBuffer data) {
        if(!this.isInitialized()) {
            Logger.INSTANCE.warning("Call of setData() on uninitialized texture %s.", this.resource);
        }
        if(this.hasData) {
            Logger.INSTANCE.warning("Call of setData() on texture %s which has already data.", this.resource);
        }
        int format;
        switch(channels) {
            case 1: format = GL11.GL_RED; break;
            case 2: format = GL30.GL_RG; break;
            case 3: format = GL11.GL_RGB; break;
            case 4: format = GL11.GL_RGBA; break;
            default:
                Logger.INSTANCE.error("Unsupported color channel count for texture %s!", this.resource);
                return;
        }        
        this.bind();
        GL11.glTexImage2D(this.target, 0, GL11.GL_RGBA8, width, height, 0, format, GL11.GL_FLOAT, data);
        GL11.glTexParameteri(this.target, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR_MIPMAP_LINEAR);
        GL11.glTexParameteri(this.target, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        GL30.glGenerateMipmap(this.target);
        this.hasData = true;
        this.width = width;
        this.height = height;
        
        this.size = 0;
        int maxLevelSize = width * height * channels;
        int dim = Math.max(width, height);
        while(dim != 0) {
            this.size += maxLevelSize;
            maxLevelSize /= 4;
            dim /= 2;
        }
    }

    boolean hasData() {
        return this.hasData;
    }
    
    boolean isInitialized() {
        return this.id != 0;
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }

    public int getSize() {
        return size;
    }
}
