/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package graphics.texture;

import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import logging.Logger;
import main.GameApp;
import org.lwjgl.BufferUtils;
import resource.ResHandle;
import resource.Resource;
import resource.loader.ImageExtraData;
import util.ImageUtil;

/**
 * Der TextureLoader ist die zentrale Klasse, um Texturen zu laden oder zu
 * erstellen.
 * @author nico3000
 */
public class TextureLoader {
    /**
     * Die Standard 2D Textur.
     */
    private static final String DEFAULT_TEXTURE_2D = "Textures/default.png";
    /**
     * Die Fehler 2D Textur.
     */
    private static final String ERROR_TEXTURE_2D = "Textures/error.png";
    
    /**
     * Liefert die Standard 2D Textur.
     * @return Die Textur
     */
    public static Texture getDefaultTexture2d() {
        return Texture.getTexture(DEFAULT_TEXTURE_2D);
    }
    
    /**
     * Liefert die Fehler 2D Textur.
     * @return Die Textur
     */
    public static Texture getErrorTexture2d() {
        return Texture.getTexture(ERROR_TEXTURE_2D);
    }
    
    /**
     * Initialisiert die Standard- und Fehlertextur.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public static boolean init() {
        Texture defTex = Texture.createTexture(Texture.Type.TEXTURE_2D, DEFAULT_TEXTURE_2D);
        if(defTex == null) {
            Logger.INSTANCE.error("DEFAULT_TEXTURE_2D could not be initialized!");
            return false;
        } else if(!loadDataFromResource(defTex)) {
            defTex.release();
            return false;
        }
        Texture errTex = Texture.createTexture(Texture.Type.TEXTURE_2D, ERROR_TEXTURE_2D);
        if(errTex == null) {
            Logger.INSTANCE.error("ERROR_TEXTURE_2D could not be initialized!");
            return false;
        } else if(!loadDataFromResource(errTex)) {
            defTex.release();
            errTex.release();
            return false;
        }
        return true;
    }
    
    /**
     * Zerstoert die Standard- und Fehlertextur.
     */
    public static void destroy() {
        Texture tex = Texture.getTexture(DEFAULT_TEXTURE_2D);
        tex.release(); // einmal fuer das vorherige getTexture()
        tex.release(); // einmal fuer das createTexture in init()
        tex = Texture.getTexture(ERROR_TEXTURE_2D);
        tex.release(); // einmal fuer das vorherige getTexture()
        tex.release(); // einmal fuer das createTexture in init()
    }
    
    /**
     * Liefert eine Resource als Textur zurueck und laedt sie dabei vorher,
     * falls es nicht schon passiert ist.
     * @param resource Die Resource
     * @return Die Textur
     */
    public static Texture getTexture(String resource) {
        Texture tex = Texture.getTexture(resource);
        if(tex != null && tex.isInitialized() && tex.hasData()) {
            return tex;
        }
        if(tex == null) {
            tex = Texture.createTexture(Texture.Type.TEXTURE_2D, resource);
            if(tex == null) {
                return getErrorTexture2d();
            }
        }
        if(!loadDataFromResource(tex)) {
            tex.release();
            return getErrorTexture2d();
        }
        return tex;
    }
    
    /**
     * Erstellt eine Textur aus einem BufferedImage.
     * @param image Das Bild
     * @return Die Textur
     */
    public static Texture createTexture(BufferedImage image) {
        Texture tex = Texture.createTexture(Texture.Type.TEXTURE_2D, "texture_buffered_image_" + image.hashCode());
        if(tex == null) {
            return getErrorTexture2d();
        }
        tex.setData(image.getWidth(), image.getHeight(), image.getColorModel().getNumComponents(), ImageUtil.bufferedImage2floatBuffer(image));
        return tex;
    }
    
    /**
     * Laedt fuer eine Textur ihre Daten aus einer Resource.
     * @param tex Die Textur
     * @param resource Die zu ladenden Daten
     * @return <tt>true</tt>, falls erfolgreich
     */
    private static boolean loadDataFromResource(Texture tex) {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(tex.getResource()));
        if(handle == null) {
            return false;
        }
        
        if(handle.getExtra() == null || !(handle.getExtra() instanceof ImageExtraData)) {
            Logger.INSTANCE.error("Invalid resource %s for texture!", tex.getResource());
            return false;
        }
        ImageExtraData extra = (ImageExtraData)handle.getExtra();
        
        FloatBuffer buffer = BufferUtils.createFloatBuffer(handle.getSize() / 4);
        ByteBuffer srcBuffer = ByteBuffer.wrap(handle.getBuffer());
        int max = extra.getWidth() * extra.getHeight() * extra.getChannels();
        for(int i=0; i < max; ++i) {
            buffer.put(srcBuffer.getFloat());
        }
        buffer.position(0);
        handle.unlock();
        
        tex.setData(extra.getWidth(), extra.getHeight(), extra.getChannels(), buffer);
        return true;
    }
}
