/*
 * This software is the proprietary information of nico3000.
 */
package gui;

import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;
import graphics.shader.Effect;
import gui.components.Component;
import gui.components.Cursor;
import gui.components.Label;
import gui.components.Window;
import input.KeyboardHandler;
import input.PointerHandler;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import logging.Logger;
import main.GameApp;
import mesh.Mesh;
import mesh.loader.generic.GenericLoader;
import org.lwjgl.BufferUtils;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.GL11;
import process.ProcessManager;
import view.ScreenElement;

/**
 *
 * @author nico3000
 */
public class Gui extends ScreenElement implements KeyboardHandler, PointerHandler {
    public static Mesh QUAD;
    public static final Effect GUI_EFFECT = new Effect();
    public static final int EFFECT_TEX_SLOT = 7;
    public static final long MAX_CLICK_MILLIS = 100;
    public static final int MAX_POINTER_BUTTONS = 8;
    
    /**
     * Initialisiert das Quad und den Effekt der GUI.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public static boolean init() {
        if(!GUI_EFFECT.initVF("Shaders/GuiVS.glsl", "Shaders/GuiFS.glsl")) {
            Logger.INSTANCE.error("Initialization of gui effect failed!");
            return false;
        }
        GUI_EFFECT.use();
        GUI_EFFECT.setUniform("g_ComponentTex", EFFECT_TEX_SLOT);
        IndexBuffer iBuffer = new IndexBuffer();
        VertexBuffer vBuffer = new VertexBuffer();
        if(!iBuffer.init() || !vBuffer.init()) {
            Logger.INSTANCE.error("Initialization of gui quad geometry failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();
            return false;
        }
        IntBuffer indices = BufferUtils.createIntBuffer(4);
        indices.put(new int[] { 0, 1, 3, 2, });
        indices.position(0);
        iBuffer.setData(indices, IndexBuffer.TRIANGLE_STRIP);
        FloatBuffer vertices = BufferUtils.createFloatBuffer(8);
        vertices.put(new float[] { 0.0f, 0.0f, 
                                   1.0f, 0.0f, 
                                   1.0f, 1.0f, 
                                   0.0f, 1.0f, });
        vertices.position(0);
        vBuffer.setData(vertices);
        vBuffer.addAttribute(VertexBuffer.ATTR_POS, 2);
        QUAD = Mesh.createMesh(new GenericLoader(vBuffer, iBuffer));
        if(QUAD == null) {
            Logger.INSTANCE.error("Initialization of gui quad failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();
            return false;
        }
        return true;
    }
    
    /**
     * Zerstoert das Quad und den Effekt der GUI.
     */
    public static void destroy() {
        QUAD.release();
        GUI_EFFECT.destroy();
    }

    /**
     * Der Name der Gui.
     */
    private final String name;
    /**
     * Die Klasse, die fuer das Look and Feel der Gui verantwortlich ist.
     */
    private final ComponentDrawer drawer;
    /**
     * Die Komponenten dieser Gui nach ihrer ID.
     */
    private Map<String,Component> components = new HashMap<>();
    /**
     * Die Komponenten nach ihrer Tiefe sortiert.
     */
    private List<Component> sortedComponents = new LinkedList<>();
    /**
     * Der aktuelle Cursor.
     */
    private final Cursor cursor = new Cursor();
    /**
     * Label fuer Tooltips.
     */
    private final Label tooltip = new Label("Empty Tooltip", true);
    /**
     * Merkt sich die Komponente unter dem Cursor.
     */
    private Component hovered = null;
    /**
     * Merkt sich die fokussierte Komponentedieser Gui.
     */
    private Component focused = null;
    /**
     * Das fokussierte Fenster.
     */
    private Window focusedWindow = null;
    /**
     * Das originale z Order des fokussierten Fensters.
     */
    private int focusedWindowOriginalDepth = 0;
    /**
     * Stati der Mausbuttons.
     */
    private long buttons[] = new long[MAX_POINTER_BUTTONS];
    /**
     * Zeitstempel der letzten Pointerbewegung.
     */
    private long timeOnHovered = 0;
    /**
     * Der ProcessManager der Gui.
     */
    private final ProcessManager processManager = new ProcessManager();

    /**
     * Konstruktor.
     * @param name Der Name der Gui
     * @param drawer Die Klasse, die fuer das Look&amp;Feel verantwortlich ist
     */
    public Gui(String name, ComponentDrawer drawer) {
        this.name = name;
        this.drawer = drawer;
        this.cursor.setId(name + "_cursor");
        this.tooltip.setId(name + "_tooltip");
        for(int i=0; i < MAX_POINTER_BUTTONS; ++i) {
            this.buttons[i] = -1;
        }
    }

    /**
     * Liefert den Namen der Gui.
     * @return Der Name der Gui
     */
    public String getName() {
        return name;
    }
    
    /**
     * Liefert eine Komponente anhand ihres Qualifiers. Dabei ist der Qualifier
     * der zusammengesetzte String aus allen Komponenten von der Wurzel bis zum
     * gesuchten Element, getrennt mit Punkten.
     * @param qualifier Der Qualifier
     * @return Die Komponente oder <tt>null</tt>, falls keine Komponente mit
     *         dem uebergebenen Qualifier existiert oder der Qualifier ungueltig
     *         ist
     */
    public Component findByQualifier(String qualifier) {
        if(qualifier == null || qualifier.trim().isEmpty()) {
            Logger.INSTANCE.warning("Qualifier must not be empty.");
            return null;
        }
        String split[] = qualifier.split("\\.", 2);
        Component comp = this.findComponentByName(split[0]);
        if(split.length == 1) {
            return comp;
        } else {
            return comp == null ? null : comp.findChildByQualifier(split[1]);
        }
    }
    
    /**
     * Liefert eine Komponente der Gui nach ihrem Namen.
     * @param name Der Name der Komponente
     * @return Die Komponente oder <tt>null</tt>, falls es keine mit dem
     *         angegebenen Namen gibt
     */
    public Component findComponentByName(String name) {
        for(Component component : this.sortedComponents) {
            if(component.getName().equals(name)) {
                return component;
            }
        }
        return null;
    }
    
    /**
     * Liefert eine Komponente anhand ihrer ID.
     * @param id Die ID der Komponente
     * @return Die Komponente oder <tt>null</tt>, falls es keine mit der
     *         angegebenen ID gibt
     */
    public Component getComponentById(String id) {
        return components.get(id);
    }
    
    /**
     * Liefert die Cursorkomponente.
     * @return Der Cursor
     */
    public Cursor getCursor() {
        return this.cursor;
    }

    /**
     * Liefert den Drawer der GUI.
     * @return Der Drawer
     */
    public ComponentDrawer getDrawer() {
        return drawer;
    }
    
    /**
     * Prueft, ob die momentan gehoverte Komponente immer noch gehovert ist und
     * setzt sie auf ungehovert, falls nötig.
     * @param x Die Cursorposition
     * @param y Die Cursorposition
     */
    private void checkHovered(int x, int y) {
        if(this.hovered != null && !this.hovered.isIn(x, y)) {
            this.hovered.setHovered(false);
            this.hovered = null;
        }
    }
    
    /**
     * Die momentan gehoverte Komponente wird ungehovert.
     */
    private void leaveHovered() {
        if(this.hovered != null) {
            this.hovered.setHovered(false);
            this.hovered = null;
        }
    }
    
    /**
     * Die momentan fokussierte Komponente wird defokussiert.
     */
    private void leaveFocused() {
        if(this.focused != null) {
            this.focused.setFocused(false);
            this.focused = null;
        }
    }
    
    private void setFocusedWindow(Window window) {
        if(this.focusedWindow != window) {
            if(this.focusedWindow != null) {
                this.focusedWindow.setLocalZOrder(this.focusedWindowOriginalDepth);
                this.focusedWindow.onFocusLost();
                this.focusedWindow = null;
            }
            this.focusedWindow = window;
            if(this.focusedWindow != null) {
                this.focusedWindowOriginalDepth = this.focusedWindow.getLocalZOrder();
                this.focusedWindow.setLocalZOrder(Integer.MAX_VALUE / 2);
                this.focusedWindow.onFocusGained();
            }
        }
    }
    
    /**
     * Fuegt der GUI eine Komponente hinzu. Die ist nur moeglich, falls sie
     * nicht bereits der GUI hinzugefuegt wurde.
     * @param comp Die neue Komponente
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean addComponent(Component comp) {
        if(comp == null) {
            Logger.INSTANCE.warning("Trying to add null as component to %s.", this);
            return false;
        } else if(comp.getId() == null) {
            Logger.INSTANCE.warning("Tried to add component %s without ID to %s.", comp, this);
            return false;
        } else if(this.components.containsKey(comp.getId())) {
            Logger.INSTANCE.warning("Trying to add second component with ID %s to %s.", comp.getId(), this);
            return false;
        } else if(!comp.init(this)) {
            Logger.INSTANCE.warning("Initialization of component %s for %s failed.", comp, this);
            return false;
        } else {
            this.components.put(comp.getId(), comp);
            this.sortedComponents.add(comp);
            return true;
        }
    }
    
    /**
     * Entfernt eine Komponente aus der GUI.
     * @param comp Die zu entfernende Komponente
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean removeComponent(Component comp) {
        if(comp == null) {
            Logger.INSTANCE.warning("Cannot remove null from %s.", this);
            return false;
        } else if(comp == this.cursor) {
            Logger.INSTANCE.warning("Cannot remove own cursor from %s. Use setVisible() instead.", this);
            return false;
        } else if(comp == this.tooltip) {
            Logger.INSTANCE.warning("Cannot remove own tooltip from %s.", this);
            return false;
        } else if(!this.components.containsKey(comp.getId())) {
            Logger.INSTANCE.warning("Tried to remove non-existing component from %s.", this);
            return false;
        } else {
            comp.destroy();
            this.components.remove(comp.getId());
            this.sortedComponents.remove(comp);
            return true;
        }
    }

    @Override
    public boolean onKeyDown(int key, char c) {
        return this.focused != null && this.focused.onKeyDown(key, c);
    }

    @Override
    public boolean onKeyUp(int key) {
        return this.focused != null && this.focused.onKeyUp(key);
    }

    @Override
    public boolean isProcessingRepeatEvents() {
        return true;
    }

    @Override
    public void setVisible(boolean visible) {
        super.setVisible(visible);
        if(this.isVisible()) {
            GameApp.getInputHandler().registerKeyboardHandler(this);
            GameApp.getInputHandler().registerPointerHandler(this);
        } else {
            GameApp.getInputHandler().unregisterKeyboardHandler(this);
            GameApp.getInputHandler().unregisterPointerHandler(this);            
        }
        if(this.cursor.isVisible()) {
            Mouse.setGrabbed(visible);
        }
    }

    @Override
    public boolean pointerMoved(int posX, int posY, int dX, int dY) {
        int oldCursorX = this.cursor.getPointerX();
        int oldCursorY = this.cursor.getPointerY();
        this.tooltip.setPosition(Component.REF_TOP | Component.REF_LEFT, posX + this.cursor.getWidth(), posY - this.cursor.getHeight());
        this.cursor.setPointerPos(posX, posY);
        this.checkHovered(posX, posY);
        int cursorDx = this.cursor.getPointerX() - oldCursorX;
        int cursorDy = this.cursor.getPointerY() - oldCursorY;
        if(cursorDx == 0 && cursorDy == 0) {
            // cursor not moved
            return true;
        }
        
        boolean consumed = false;
        ListIterator<Component> iter = this.sortedComponents.listIterator(this.sortedComponents.size());
        while(!consumed && iter.hasPrevious()) {
            Component comp = iter.previous();
            if(comp == this.hovered) {
                // keine andere komponente ueber der bereits gehoverten
                consumed = true;
            } else if(comp.isVisible() && !comp.isGhost() && comp.isIn(this.cursor.getPointerX(), this.cursor.getPointerY())) {
                // andere komponente (sichtbar, kein geist) ueber der bereits gehoverten
                this.leaveHovered();
                this.hovered = comp;
                this.hovered.setHovered(true);
                this.timeOnHovered = 0;
                if(comp.getTooltip() != null) {
                    this.tooltip.setSize(12 * comp.getTooltip().length(), 24);
                }
                this.tooltip.setText(comp.getTooltip());
                consumed = true;
            }
        }
        if(this.focused != null) {
            this.focused.onCursorMoved(this.cursor.getPointerX(), this.cursor.getPointerY(), cursorDx, cursorDy);
        }
        if(this.hovered != null && this.hovered != this.focused) {
            // TODO: dX dY bei frisch gehoverten komponenten evtl. falsch
            this.hovered.onCursorMoved(this.cursor.getPointerX(), this.cursor.getPointerY(), cursorDx, cursorDy);
        }
        return this.isDragging();
    }

    @Override
    public boolean onButtonDown(int button) {
        ListIterator<Component> iter = this.sortedComponents.listIterator(this.components.size());
        boolean hit = false; // gibt an, ob eine komponente getroffen wurde
        while(!hit && iter.hasPrevious()) {
            Component comp = iter.previous();
            if(comp.isVisible() && !comp.isGhost() && comp.isIn(this.cursor.getPointerX(), this.cursor.getPointerY())) {
                // komponente sichtbar, kein geist und wurde getroffen
                hit = true;
                if(this.focused != comp) {
                    // komponente hat noch nicht den fokus
                    this.leaveFocused();
                    this.focused = comp;
                    this.focused.setFocused(true);
                }
            }
        }
        if(!hit) {
            // es wurde nichts getroffen
            this.leaveFocused();
            this.setFocusedWindow(null);
            return false;
        } else {
            // es wurde etwas getroffen
            this.buttons[button] = GameApp.getRealTimer().getTimeStamp();
            this.focused.onButtonDown(button, this.cursor.getPointerX(), this.cursor.getPointerY());
            this.setFocusedWindow(this.focused.getParentWindow());
            return true;
        }
    }

    @Override
    public boolean onButtonUp(int button) {
        if(this.focused != null) {
            boolean clicked = GameApp.getRealTimer().getTimeStamp() - this.buttons[button] < MAX_CLICK_MILLIS;
            this.buttons[button] = -1;
            this.focused.onButtonUp(button, this.cursor.getPointerX(), this.cursor.getPointerY(), clicked);
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Liefert, ob gerade ein Button der Maus gedrueckt ist.
     * @return <tt>true</tt>, falls ein Button gedrueckt ist
     */
    private boolean isDragging() {
        return this.buttons[0] != -1 ||
                this.buttons[1] != -1 ||
                this.buttons[2] != -1 ||
                this.buttons[3] != -1 ||
                this.buttons[4] != -1 ||
                this.buttons[5] != -1 ||
                this.buttons[6] != -1 ||
                this.buttons[7] != -1;
    }

    @Override
    public boolean onWheelMoved(int delta) {
        if(this.hovered != null) {
            return this.hovered.onWheelMoved(delta);
        } else {
            return false;
        }
    }

    @Override
    public boolean onRestore() {
        this.setVisible(true);
        if(!this.addComponent(this.cursor)) {
            return false;
        }
        if(!this.addComponent(this.tooltip)) {
            this.onDestroy();
            return false;
        }
        this.tooltip.setSize(192, 24);
        return true;
    }

    @Override
    public void onDestroy() {
        for(Component comp : this.sortedComponents) {
            comp.destroy();
        }
        this.sortedComponents.clear();
        this.components.clear();
    }

    @Override
    public boolean onLostDevice() {
        return true;
    }

    @Override
    public void onRender() {
        GL11.glDisable(GL11.GL_DEPTH_TEST);
        for(Component comp : this.sortedComponents) {
            if(comp.isVisible()) {
                comp.onRender();
            }
        }
        GL11.glEnable(GL11.GL_DEPTH_TEST);
    }

    @Override
    public void onUpdate(long deltaMillis) {
        this.timeOnHovered += deltaMillis;
        if(this.timeOnHovered < 500 || this.hovered == null || this.tooltip.getText() == null) {
            this.tooltip.setVisible(false);
        } else {
            this.tooltip.setVisible(true);
        }
        boolean sorted = true;
        int lastZ = Integer.MIN_VALUE;
        int currentZ;
        for(Component comp : this.sortedComponents) {
            if(comp.isVisible()) {
                comp.onUpdate(deltaMillis);
                currentZ = comp.getGlobalZOrder();
                if(lastZ > comp.getGlobalZOrder()) {
                    sorted = false;
                }
                lastZ = currentZ;
            }
        }
        if(!sorted) {
            Logger.INSTANCE.info("Sorting components of %s.", this);
            Collections.sort(this.sortedComponents);
        }
        this.processManager.update(deltaMillis);
    }

    /**
     * Liefert den ProcessManager der Gui.
     * @return Der ProcessManager
     */
    public ProcessManager getProcessManager() {
        return this.processManager;
    }
    
    @Override
    public String toString() {
        return String.format("Gui '%s'", this.getName());
    }
}
