/*
 * This software is the proprietary information of nico3000.
 */
package gui;

import event.EventData;
import event.events.StringEvent;
import event.events.ToggleEvent;
import gui.components.Component;
import logging.Logger;
import main.GameApp;

/**
 * Eine GuiAction kann mit einer Komponente verknuepft werden. Dazu kann direkt
 * ein Event angegeben werden, welches geworfen werden soll. Alternativ kann
 * eine Action mittels Miniscript(tm) (solange wir noch keine anstaendige
 * Luaanbindung haben :P) angesteuert werden.
 * @author nico3000
 */
public abstract class GuiAction {
    /**
     * Erzeugt eine GuiAction aus einem String.
     * @param actionString Der String
     * @return Die erzeugte Action bei Erfolg; sonst <tt>null</tt>
     */
    public static GuiAction fromString(String actionString) {
        if(actionString == null) {
            return null;
        } else if(actionString.startsWith("miniscript:")) {
            int splitIndex = actionString.lastIndexOf(".");
            if(splitIndex > 0) {
                String identifier = actionString.substring(11, splitIndex);
                String method = actionString.substring(splitIndex + 1);
                MiniscriptAction action = new MiniscriptAction();
                if(action.setAction(identifier, method)) {
                    return action;
                }
            }
            Logger.INSTANCE.warning("Illegal miniscript \"%s\"", actionString.substring(11));
            return null;
        } else {
            EventAction action = new EventAction();
            return action.setAction(actionString) ? action : null;
        }
    }
    
    /**
     * Privater Konstruktor um unkntrolliertes Vererben zu verhindern.
     */
    private GuiAction() {}
    
    /**
     * Feuert die Action mit optionalen Daten ab. Dabei haengt es von der
     * speziellen GuiAction ab, wie die Daten benutzt werden.
     * @param optionalData Die optionalen Daten
     */
    public abstract void fire(Object optionalData);
    
    /**
     * Eine EventAction feuert ein ihr zugewiesenes Event ab.
     */
    public static class EventAction extends GuiAction {
        /**
         * Die Eventklasse.
         */
        private Class<?> action;
        
        /**
         * Bestimmt das auszuloesende Event aus einem String.
         * @param eventClass Das auszuloesende Event
         * @return <tt>true</tt>, falls die Eventklasse existiert
         */
        private boolean setAction(String eventClass) {
            try {
                this.action = Class.forName(eventClass);
                return true;
            } catch (ClassNotFoundException ex) {
                Logger.INSTANCE.error(ex.getMessage());
                return false;
            }
        }
        
        @Override
        public void fire(Object data) {
            try {
                EventData evt = (EventData)this.action.newInstance();
                if(evt instanceof StringEvent && data instanceof String) {
                    ((StringEvent)evt).setString((String)data);
                } else if(evt instanceof ToggleEvent && data instanceof Boolean) {
                    ((ToggleEvent)evt).setActivated((Boolean)data);
                }
                GameApp.getEventManager().queueEvent(evt);
            } catch (InstantiationException | IllegalAccessException | ClassCastException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
        }
    }
    
    /**
     * Eine MiniscriptAction kann momentan eine Komponente einer Gui sichtbar
     * oder unsichtbar machen.
     */
    public static class MiniscriptAction extends GuiAction {
        /**
         * Die Methode, die aufgerufen werden soll.
         */
        private static enum Method {
            /**
             * Verteckt eine Komponente.
             */
            HIDE,
            /**
             * Macht eine Komponente sichtbar.
             */
            SHOW,
            /**
             * Schaltet zwischen sichtbar und versteckt um.
             */
            TOGGLE,
        }
        
        /**
         * Die mit dieser Action verknuepfte Methode.
         */
        private Method method;
        /**
         * Die ID der Komponente.
         */
        private String identifier;
        
        /**
         * Setzt die Daten dieser Action.
         * @param identifier Die ID der Komponente
         * @param method Die Methode
         * @return <tt>true</tt>, wenn die Methode gueltig ist
         */
        private boolean setAction(String identifier, String method) {
            this.identifier = identifier;
            int braceIndex = method.indexOf('(');
            if(braceIndex == -1) {
                return false;
            }
            method = method.substring(0, braceIndex);
            switch(method) {
                case "hide":
                    this.method = Method.HIDE;
                    break;
                case "show":
                    this.method = Method.SHOW;
                    break;
                case "toggle":
                    this.method = Method.TOGGLE;
                    break;
                default:
                    return false;
            }
            return true;
        }
        
        @Override
        public void fire(Object data) {
            Component comp = GameApp.getGameLogic().getHumanView().getGuiComponent(this.identifier);
            if(comp != null) {
                switch(this.method) {
                    case HIDE:
                        comp.hide();
                        break;
                    case SHOW:
                        comp.show();
                        break;
                    case TOGGLE:
                        if(comp.isVisible()) {
                            comp.hide();
                        } else {
                            comp.show();
                        }
                        break;
                }
            }
        }
    }
}
