/*
 * This software is the proprietary information of nico3000.
 */
package gui.components;

import graphics.shader.UniformBuffer;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;
import gui.Gui;
import gui.GuiAction;
import java.awt.Rectangle;
import java.nio.FloatBuffer;
import java.util.HashSet;
import java.util.Set;
import logging.Logger;
import main.GameApp;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector4f;
import process.EngineProcess;
import util.MathUtil;

/**
 * Repraesentiert ein allgemeines Gui Element. Es besteht aus einem
 * positionierten Rechteck gefuellt mit einer Textur.
 * @author nico3000
 */
public abstract class Component implements Comparable<Component> {
    /**
     * Untere Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_BOTTOM = 1;     // ...0001
    /**
     * Vertikale Mitte der Komponente als Bezugspunkt fuer
     * <tt>setPosition()</tt>.
     */
    public static final int REF_VCENTER = 3;    // ...0011
    /**
     * Obere Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_TOP = 2;        // ...0010
    /**
     * Linke Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_LEFT = 4;       // ...0100
    /**
     * Horizontale Mitte der Komponente als Bezugspunkt fuer
     * <tt>setPosition()</tt>.
     */
    public static final int REF_HCENTER = 12;   // ...1100
    /**
     * Rechte Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_RIGHT = 8;      // ...1000
    /**
     * Maske des vertikalen Bezugspunktes der Komponente fuer
     * <tt>setPosition()</tt>.
     */
    private static final int REF_VER_MASK = 3;  // ...0011
    /**
     * Maske des horizontalen Bezugspunktes der Komponente fuer
     * <tt>setPosition()</tt>.
     */
    private static final int REF_HOR_MASK = 12; // ...1100
    
    /**
     * Kann der Methode <tt>setPosition</tt> uebergeben werden, wenn nur eine
     * Koordinate geaendert werden soll.
     */
    public static final int POS_NO_CHANGE_INT = Integer.MIN_VALUE;
    /**
     * Kann der Methode <tt>setPosition</tt> uebergeben werden, wenn nur eine
     * Koordinate geaendert werden soll.
     */
    public static final float POS_NO_CHANGE_FLOAT = Float.NEGATIVE_INFINITY;
    protected static final int REPAINT_ON_RESIZE = 1;
    protected static final int REPAINT_ON_MOVE = 2;
    protected static final int REPAINT_ON_HOVER = 4;
    protected static final int REPAINT_ON_BLUR = 8;
    protected static final int REPAINT_ON_FOCUS_GAINED = 16;
    protected static final int REPAINT_ON_FOCUS_LOST = 32;
    protected static final int REPAINT_ON_NULL_TEXTURE = 64;
    
   
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */
    private static final Matrix4f TEMP_SCALE = new Matrix4f();
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */
    private static final Matrix4f TEMP_TRANS = new Matrix4f();
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */    
    private static final Matrix4f TEMP_TRANSFORM = new Matrix4f();
    /**
     * Zaehler der Komponenten.
     */
    private static int counter = 0;
    
    /**
     * Der Uniformbuffer der Komponente.
     */
    private final UniformBuffer transform = new UniformBuffer();
    /**
     * Die Daten des Uniformbuffers.
     */
    private final FloatBuffer transformBuffer = BufferUtils.createFloatBuffer(21);
    /**
     * Gibt an, ob der Buffer aktualisiert werden muss.
     */
    private boolean componentChanged = false;
    /**
     * Die Textur der Komponente.
     */
    private Texture texture = TextureLoader.getDefaultTexture2d();
    /**
     * Die aktuelle Ebene der Komponente.
     */
    private int zOrder = 0;
    /**
     * Gibt an, ob die Komponente sichtbar ist.
     */
    private boolean isVisible = true;
    /**
     * Gibt an, ob die Komponente neu gezeichnet werden muss.
     */
    private boolean needsRepaint = false;
    /**
     * Die Position und Groesse der Komponente als Rechteck in Pixelkoordinaten.
     */
    private final Rectangle rect = new Rectangle();
    /**
     * Die Gui dieser Komponente.
     */
    private Gui gui;
    /**
     * Die ID der Komponente.
     */
    private String id;
    /**
     * Der Name der Komponente.
     */
    private String name;
    /**
     * Gibt an, wann die Komponente neu gezeichnet werden muss.
     */
    private int repaintOn;
    /**
     * Tooltip der Komponente.
     */
    private String tooltip = null;
    /**
     * Die Komponente, die diese Komponente enthaelt, sofern vorhanden.
     */
    private Component container = null;
    /**
     * Menge aller Komponenten, die in dieser enthalten sind.
     */
    private final Set<Component> children = new HashSet<>();
    /**
     * Repraesntiert das Clippingrechteck dieser Komponente im projection space.
     */
    private final Vector4f clippingRect = new Vector4f(-1.0f, -1.0f, +1.0f, +1.0f);
    /**
     * Eine mit dieser Komponente verknuepfte Action.
     */
    private GuiAction action = null;
    /**
     * Die Transparenz der Komponente (0.0 entpricht komplett transparent).
     */
    private float transparency = 1.0f;
    /**
     * Gibt an, ob die Komponente den Fokus hat.
     */
    private boolean focused = false;
    /**
     * Gibt an, ob sich der Cursor ueber der Komponente befindet.
     */
    private boolean hovered = false;
    
    /**
     * Konstruktor.
     */
    protected Component(int repaintOn) {
        this.name = "Component_" + counter++;
        this.repaintOn = repaintOn;
    }

    /**
     * Setzt den Namen der Komponente.
     * @param name Der neue Name
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Liefert den Namen der Komponente.
     * @return Der Name
     */
    public String getName() {
        return this.name;
    }

    /**
     * Liefert die ID der Komponente.
     * @return Die ID
     */
    public String getId() {
        return id;
    }

    /**
     * Setzt die ID dieser Komponente. Dieser Aufruf funktioniert nur einmal.
     * Wird <tt>null</tt> uebergeben, wird der Hashcode als ID verwendet.
     * @param id Die ID
     */
    public void setId(String id) {
        if(this.id != null) {
            Logger.INSTANCE.warning("ID of %s already set.", this);
        } else if(id == null) {
            this.id = "" + this.hashCode();
        } else {
            this.id = id;
        }
    }
    
    /**
     * Liefert den Qualifier dieser Komponente. Dieser setzt sich zusammen aus
     * dem Qualifier des Containers und dem Namen der Komponente, getrennt durch
     * einen Punkt. Gibt es keinen Container, wird der Name der zugehoerigen Gui
     * verwendet.
     * @return Der Qualifier
     */
    public String getQualifier() {
        if(this.container != null) {
            return String.format("%s.%s", this.container.getQualifier(), this.name);
        } else {
            return String.format("%s.%s", this.gui == null ? "<no gui>" : this.gui.getName(), this.name);
        }
    }
    
    /**
     * Setzt ein Repaint Bit.
     * @param condition Das Bit
     */
    protected void setRepaintBit(int condition) {
        this.repaintOn |= condition;
    }
    
    /**
     * Loescht ein Repaint Bit.
     * @param condition Das Bit
     */
    protected void clearRepaintBit(int condition) {
        this.repaintOn &= ~condition;
    }

    /**
     * Setzt einen Tooltip fuer diese Komponente.
     * @param tooltip Der neue Tooltip
     */
    public void setTooltip(String tooltip) {
        this.tooltip = tooltip;
    }

    /**
     * Liefert den Tooltip fuer diese Komponente.
     * @return Der Tooltip
     */
    public String getTooltip() {
        return this.tooltip;
    }

    /**
     * Setzt eine neue Action fuer diese Komponente. Wie die Action aktiviert
     * wird, haengt von der Komponente ab.
     * @param action Die neue Action
     */
    public void setAction(GuiAction action) {
        this.action = action;
    }
    
    /**
     * Markiert die Komponente als neu zu zeichnend, falls eine bestimmte
     * Bedingung erfuellt ist. Der naechste Rendervorgang wird dann vorher
     * <tt>repaint()</tt> aufrufen.
     * @param condition Die Bedingung
     */
    private void setNeedsRepaintIf(int condition) {
        if((this.repaintOn & condition) == condition) {
            this.needsRepaint = true;
        }
    }
    
    /**
     * Feuer die mit dieser ID verknuepfte Action ab.
     * @param optionalData Optionale Zusatzdaten, die von der speziellen Action
     *                     evtl. genutzt werden
     */
    protected void fireAction(Object optionalData) {
        if(this.action != null) {
            this.action.fire(optionalData);
        }
    }
    
    /**
     * Entspricht einem Aufruf von <tt>fireAction(null)<tt>.
     */
    protected void fireAction() {
        this.fireAction(null);
    }
        
    /**
     * Liefert die Gui dieser Komponente.
     * @return Die Gui
     */
    protected Gui getGui() {
        return this.gui;
    }
    
    /**
     * Setzt die lokale Tiefe der Komponente bezueglich des Containers 
     * (initial auf <tt>0</tt>). Die Komponenten werden in aufsteigender
     * Reihenfolge der globalen Tiefe gerendert. Die globale Tiefe ist die Summe
     * der globalen Tiefe des Containers, sofern vorhanden, und der lokalen
     * Tiefe der Komponente.
     * @param zOrder Die neue lokale Tiefe
     */
    public void setLocalZOrder(int zOrder) {
        this.zOrder = zOrder;
    }
    
    /**
     * Liefert die lokale Tiefe.
     * @return Die lokale Tiefe
     */
    public int getLocalZOrder() {
        return this.zOrder;
    }
    
    /**
     * Liefert die globale Tiefe der Komponente.
     * @return Die globale Tiefe
     */
    public int getGlobalZOrder() {
        return this.container == null ? this.zOrder : this.zOrder + this.container.getGlobalZOrder();
    }
    
    /**
     * Liefert die Sichtbarkeit der Komponente.
     * @return <tt>true</tt>, falls sichtbar
     */
    public boolean isVisible() {
        return (this.container == null || this.container.isVisible()) && this.isVisible;
    }
    
    /**
     * Setzt die Sichtbarkeit der Komponente.
     * @param visible Die neue Sichtbarkeit
     */
    public void setVisible(boolean visible) {
        this.isVisible = visible;
        if (this.isVisible) {
        	this.needsRepaint();
        }
    }

    @Override
    public int compareTo(Component other) {
        return other == null ? -1 : Integer.compare(this.getGlobalZOrder(), other.getGlobalZOrder());
    }
    
    /**
     * Setzt den Hoverzustand der Komponente.
     * @param hovered Der neue Zustand
     */
    public void setHovered(boolean hovered) {
        if(this.hovered != hovered) {
            this.hovered = hovered;
            if(this.hovered) {
                this.onHover();
                this.setNeedsRepaintIf(REPAINT_ON_HOVER);
            } else {
                this.onBlur();
                this.setNeedsRepaintIf(REPAINT_ON_BLUR);
            }
        }
    }
    
    /**
     * Setzt den Fokuszustand der Komponente.
     * @param focused Der neue Zustand
     */
    public void setFocused(boolean focused) {
        if(this.focused != focused) {
            this.focused = focused;
            if(this.focused) {
                this.onFocusGained();
                this.setNeedsRepaintIf(REPAINT_ON_FOCUS_GAINED);
            } else {
                this.onFocusLost();
                this.setNeedsRepaintIf(REPAINT_ON_FOCUS_LOST);
            }
        }
    }

    /**
     * Liefert den Fokuszustand.
     * @return Der Zustand
     */
    public boolean isFocused() {
        return this.focused;
    }

    /**
     * Liefert den Hoverzustand.
     * @return Der Zustand
     */
    public boolean isHovered() {
        return this.hovered;
    }
    
    /**
     * Fuegt dieser Komponente eine Komponente als Kind hinzu und entfernt sie
     * gleichzeitig von ihrem aktuellen Container, sofern vorhanden. Der Aufruf
     * schlaegt fehl, falls die hinzuzufuegende Komponente ein Vorgaenger dieser
     * Komponente oder bereits ein Kind ist oder beide Komponenten nicht Teil
     * derselben Gui sind.
     * @param child Die neue Kindkomponente
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean addChild(Component child) {
        if(this.gui != child.gui) {
            Logger.INSTANCE.warning("Tried to add component (%s) to a component (%s) of another gui.", this.getName(), child.getName());
            return false;
        }
        if(child instanceof Window) {
            Logger.INSTANCE.warning("Windows (%s) can't be added as a child component.", child.getName());
            return false;
        }
        Component predecessor = this;
        while(predecessor != null) {
            if(predecessor == child) {
                Logger.INSTANCE.error("Tried to add a predecessor (%s) of a node (%s) as its child.", child.getName(), this.getName());
                return false;
            }
            predecessor = child.container;
        }
        if(!this.children.add(child)) {
            Logger.INSTANCE.error("Tried to add a component (%s) as a child of another component (%s) twice.", child.getName(), this.getName());
            return false;
        } else {
            if(child.container != null) {
                child.container.children.remove(child);
            }
            child.container = this;
            child.componentChanged = true;
            return true;
        }
    }
    
    /**
     * Gibt an, ob die Komponente gehovert oder fokussiert werden kann.
     * @return <tt>true</tt>, falls die Komponente gehovert und fokussiert
     *         werden kann; <tt>false</tt> sonst
     */
    public boolean isGhost() {
        return false;
    }
    
    /**
     * Entspricht einem Aufruf von <tt>setSize(width, height)</tt> und
     * anschliessendem <tt>setPosition(x, y)</tt>.
     * @param x Der Abstand vom linken (positiv) oder rechten (negativ) 
     *          Containerrand
     * @param y Der Abstand vom unteren (positiv) oder oberen (negativ)
     *          Containerrand
     * @param width Die Breite
     * @param height Die Hoehe
     */
    public void setDimension(int x, int y, int width, int height) {
        this.setSize(width, height);
        this.setPosition(x, y);
    }
    
    /**
     * Setzt die Position des Mittelpunkts der Komponente relativ zum Container.
     * @param percentageX Der relative Abstand vom linken (positiv) bzw. rechten
     *                    (negativ) Fenster- bzw. Containerrand
     * @param percentageY Der relative Abstand vom oberen (positiv) bzw. unteren
     *                    (negativ) Fenster- bzw. Containerrand
     */
    public void setPosition(float percentageX, float percentageY) {
        int x = (int)(percentageX * (float)this.getContainerWidth());
        int y = (int)(percentageY * (float)this.getContainerHeight());
        this.setPosition(REF_HCENTER | REF_VCENTER,
                Float.isInfinite(percentageX) ? POS_NO_CHANGE_INT : x,
                Float.isInfinite(percentageY) ? POS_NO_CHANGE_INT : y);
    }
    
    /**
     * Setzt die Positionen relativ zum Fensterrand. Positive Werte werden als
     * Abstand vom linken bzw. unteren Rand interpretiert, waehrend negative
     * Werte als Abstand vom rechten, bzw. oberen Rand interpretiert werden.
     * Nach einem Aufruf von <tt>setSize()</tt> muessen diese Werte evtl. neu
     * gesetzt werden. Der Nullpunkt ist dabei der linke untere Rand des
     * Fensters bzw. Containers.
     * @param x Der Abstand vom linken (positiv) oder rechten (negativ) Fenster-
     *          bzw. Containerrand
     * @param y Der Abstand vom unteren (positiv) oder oberen (negativ) Fenster-
     *          bzw. Containerrand
     */
    public void setPosition(int x, int y) {
        int ref = (x < 0 ? REF_RIGHT : REF_LEFT) | (y < 0 ? REF_TOP : REF_BOTTOM);
        this.setPosition(ref, x, y);
    }
    
    /**
     * Setzt die Position der Komponente anhand eines Referenzpunktes. Der
     * Nullpunkt ist dabei der linke untere Rand des Fensters bzw. Containers.
     * @param reference Der Referenzpunkt der Komponente; eine Kombination aus
     *                  <tt>REF_BOTTOM</tt>, <tt>REF_VCENTER</tt>,
     *                  <tt>REF_TOP</tt>, <tt>REF_LEFT</tt>,
     *                  <tt>REF_HCENTER</tt> und <tt>REF_RIGHT</tt>
     * @param x Die neue x-Koordinate des Referenzpunktes
     * @param y Die neue y-Koordinate des Referenzpunktes 
     */
    public void setPosition(int reference, int x, int y) {
        int positiveX = x < 0 ? x + this.getContainerWidth() : x;
        int positiveY = y < 0 ? y + this.getContainerHeight(): y;     
        switch(reference & REF_VER_MASK) {
            case REF_TOP:
                positiveY -= this.rect.height;
                break;
            case REF_VCENTER:
                positiveY -= this.rect.height / 2;
                break;
        }
        switch(reference & REF_HOR_MASK) {
            case REF_RIGHT:
                positiveX -= this.rect.width;
                break;
            case REF_HCENTER:
                positiveX -= this.rect.width / 2;
                break;
        }
        this.setAbsolutePosition(
                x == POS_NO_CHANGE_INT ? POS_NO_CHANGE_INT : positiveX,
                y == POS_NO_CHANGE_INT ? POS_NO_CHANGE_INT : positiveY);
    }
    
    /**
     * Setzt die Position der linken unteren Ecke der Komponente. Der Nullpunkt
     * ist dabei der linke untere Rand des Fensters oder, sofern vorhanden, des
     * Containers.
     * @param x Die neue x-Koordinate
     * @param y Die neue y-Koordinate
     */
    public void setAbsolutePosition(int x, int y) {
        if(x != POS_NO_CHANGE_INT) {
            this.rect.x = x;
        }
        if(y != POS_NO_CHANGE_INT) {
            this.rect.y = y;
        }
        this.componentChanged = true;
        this.setNeedsRepaintIf(REPAINT_ON_MOVE);
    }
    
    /**
     * Setzt die Groesse der Komponente in Pixel.
     * @param width Die Breite
     * @param height Die Hoehe
     */
    public void setSize(int width, int height) {
        this.rect.width = width;
        this.rect.height = height;
        this.componentChanged = true;
        if(width > 0 && height > 0) {
            this.setNeedsRepaintIf(REPAINT_ON_RESIZE);
        }
    }
    
    /**
     * Setzt die eigene Groesse prozentual zur Fenster- / Containergroesse.
     * @param percentageWidth Die Breite
     * @param percentageHeight Die Hoehe
     */
    public void setSize(float percentageWidth, float percentageHeight) {
        this.setSize(this.relToAbsWidth(percentageWidth), this.relToAbsHeight(percentageHeight));
    }
    
    /**
     * Wandelt eine prozentuale Breitenangabe fuer diese Komponente in eine
     * absolute um.
     * @param percentageWidth Breite in Prozent
     * @return Breite in Pixeln
     */
    public int relToAbsWidth(float percentageWidth) {
        return (int)((float)this.getContainerWidth() * percentageWidth);
    }
    
    /**
     * Wandelt eine prozentuale Hoehenangabe fuer diese Komponente in eine
     * absolute um.
     * @param percentageHeight Hoehe in Prozent
     * @return Hoehe in Pixeln
     */
    public int relToAbsHeight(float percentageHeight) {
        return (int)((float)this.getContainerHeight() * percentageHeight);
    }
    
    /**
     * Liefert die Breite in Pixel der Komponte.
     * @return Die Breite
     */
    public int getWidth() {
        return this.rect.width;
    }
    
    /**
     * Liefert die Hoehe in Pixel der Komponte.
     * @return Die Hoehe
     */
    public int getHeight() {
        return this.rect.height;
    }
    
    /**
     * Existiert eine Elternkomponente wird dessen Breite zurueckgegeben,
     * ansonsten Breite des Fensters.
     * @return Die Breite des Containers
     */
    private int getContainerWidth() {
        return this.container == null ? GameApp.getGraphicsLayer().getWidth() : this.container.getWidth();
    }
    
    /**
     * Existiert eine Elternkomponente wird dessen Hoehe zurueckgegeben,
     * ansonsten Hoehe des Fensters.
     * @return Die Hoehe des Containers
     */
    private int getContainerHeight() {
        return this.container == null ? GameApp.getGraphicsLayer().getHeight(): this.container.getHeight();
    }
    
    /**
     * Liefert die globale x-Koordinate der linken unteren Ecke dieser
     * Komponente. Der Nullpunkt ist dabei die linke untere Ecke des Fensters.
     * @return Die x-Koordinate
     */
    public int getGlobalX() {
        if(this.container == null) {
            return this.rect.x;
        } else {
            return this.rect.x + this.container.getGlobalX();
        }
    }
    
    /**
     * Liefert die globale y-Koordinate der linken unteren Ecke dieser
     * Komponente. Der Nullpunkt ist dabei die linke untere Ecke des Fensters.
     * @return Die y-Koordinate
     */
    public int getGlobalY() {
        if(this.container == null) {
            return this.rect.y;
        } else {
            return this.rect.y + this.container.getGlobalY();
        }
    }
    
    /**
     * Wird aufgerufen, sobald sich der Cursor sich ueber die Komponente bewegt
     * hat.
     */
    public void onHover() {}
    
    /**
     * Wird aufgerufen, sobald der Cursor nicht mehr ueber der Komponente ist.
     */  
    public void onBlur() {}
    
    /**
     * Wird aufgerufen, wenn die Komponente gehovert oder fokussiert ist und
     * sich der Cursor bewegt hat.
     * @param x Die neue absolute x Position des Cursors
     * @param y Die neue absolute y Position des Cursors
     * @param dx Die Bewegung des Cursors in x Richtung
     * @param dy Die Bewegung des Cursors in y Richtung
     */
    public void onCursorMoved(int posX, int posY, int dX, int dY) {}
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus bekommen soll.
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */    
    public void onFocusGained() {}
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus verliert.
     */  
    public void onFocusLost() {}
    
    /**
     * Wird aufgerufen, wenn eine Taste des Cursors auf der Komponente gedrueckt
     * wird.
     * @param button Die Taste
     * @param x Die x-Koordinate des Cursors
     * @param y Die y-Koordinate des Cursors
     */
    public void onButtonDown(int button, int x, int y) {}
    
    /**
     * Wird aufgerufen, wenn eine Taste des Cursors auf der Komponente
     * losgelassen wird.
     * @param button Die Taste
     * @param x Die x-Koordinate des Cursors
     * @param y Die y-Koordinate des Cursors
     * @param clicked <tt>true</tt>, falls die Zeit, in der die Taste unten war,
     *                einen bestimmten Wert nicht ueberschreitet und somit als
     *                Klick interpretiert werden kann
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */
    public void onButtonUp(int button, int x, int y, boolean clicked) {}
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus hat und eine Taste der
     * Tastatur gedrueckt wurde.
     * @param key Die Taste
     * @param c Der zugehoerige Character, falls vorhanden; sonst <tt>0</tt>
     * @return <tt>true</tt>, falls die Komponente die Aktion akzeptiert
     */
    public boolean onKeyDown(int key, char c) { return false; }
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus hat und eine Taste der
     * Tastatur losgelassen wurde.
     * @param key Die Taste
     * @return <tt>true</tt>, falls die Komponente die Aktion akzeptiert
     */
    public boolean onKeyUp(int key) { return false; }
    
    /**
     * Wird aufgerufen, wenn sich der Cursor auf der Komponente befindet und das
     * Mausrad gedreht wird.
     * @param delta Die Anzahl Einheiten der Drehung
     * @return <tt>true</tt>, falls die Komponente die Aktion akzeptiert
     */
    public boolean onWheelMoved(int delta) { return false; }

    /**
     * Wird waehrend jedem Frame aufgerufen.
     * @param deltaMillis Die Dauer des letzten Frames in Milliskunden
     */
    public void onUpdate(long deltaMillis) {}
    
    /**
     * Prueft, ob ein Punkt innerhalb der Komponente liegt.
     * @param x Die x-Koordinate
     * @param y Die y-Koordinate
     * @return <tt>true</tt>, falls innerhalb
     */
    public boolean isIn(int x, int y) {
        if(this.container == null) {
            return this.rect.contains(x, y);
        } else {
            return this.container.isIn(x, y) && this.rect.contains(x - this.container.getGlobalX(), y - this.container.getGlobalY());
        }
    }
    
    /**
     * Zerstoert die Komponente. Sollte nnur von der Gui direkt aufgerufen
     * werden.
     */
    public void destroy() {
        if(this.container != null) {
            this.container.children.remove(this);
        }
        this.transform.destroy();
        if(this.texture != null) {
            this.texture.release();
        }
    }

    /**
     * Setzt die Textur der Komponente.
     * @param texture Die neue Textur
     */
    public void setTexture(Texture texture) {
        if(this.texture != null) {
            this.texture.release();
        }
        this.texture = texture;
    }
    
    /**
     * Setzt die Textur der Komponente.
     * @param texture Die neue Textur
     */
    public void setTextureWithoutRelease(Texture texture) {
    	this.texture = texture;
    }
    
    /**
     * Mit dieser Methode sollte eine Komponente ihre Textur aktualisieren. Die
     * Methode sollte nur innerhalb der Klasse <tt>Component.java</tt>
     * aufgerufen werden.
     */
    protected abstract void repaint();
    
    /**
     * Setzt die lokale Transparenz.
     * @param transparency Die neue lokale Transparenz
     */
    public void setTransparency(float transparency) {
        this.transparency = transparency;
        this.componentChanged = true;
    }
    
    /**
     * Liefert die globale Transparenz. Diese ist das Produkt aus der lokalen
     * Transparenz und der globalen Transparenz des Containers.
     * @return Die globale Transparenz
     */
    private float getTransparency() {
        return this.container == null ? this.transparency : this.container.getTransparency() * this.transparency;
    }
    
    /**
     * Aktualisiert den Buffer der Komponente.
     */
    private void updateBuffer() {
        float width = (float)GameApp.getGraphicsLayer().getWidth();
        float height = (float)GameApp.getGraphicsLayer().getHeight();
        float gPosX = -1.0f + 2.0f * (float)this.getGlobalX() / width;
        float gPosY = -1.0f + 2.0f * (float)this.getGlobalY() / height;
        float gWidth = 2.0f * (float)this.getWidth() / width;
        float gHeight = 2.0f * (float)this.getHeight() / height;
        
        MathUtil.scaling(Component.TEMP_SCALE, gWidth, gHeight, 1.0f);
        MathUtil.translation(Component.TEMP_TRANS, gPosX, gPosY, 0.0f);
        MathUtil.mul(Component.TEMP_TRANSFORM, Component.TEMP_TRANS, Component.TEMP_SCALE);
        this.transformBuffer.position(0);
        Component.TEMP_TRANSFORM.store(this.transformBuffer);
        
        gPosX += 2.0f / width;
        gPosY += 2.0f / height;
        gWidth -= 4.0f / width;
        gHeight -= 4.0f / height;
        this.clippingRect.set(gPosX, gPosY, gPosX + gWidth, gPosY + gHeight);

        if(this.container != null) {
            this.container.clippingRect.store(this.transformBuffer);
        } else {
            this.transformBuffer.put(new float[] { -1.0f, -1.0f, +1.0f, +1.0f, });
        }
        
        this.transformBuffer.put(this.getTransparency());
        
        this.transformBuffer.position(0);
        this.transform.setData(this.transformBuffer);
        
        for(Component child : this.children) {
            child.componentChanged = true;
        }
    }

    /**
     * Initialisiert die Komponente.
     * @param gui Die Gui
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init(Gui gui) {
        this.gui = gui;
        this.transform.setBindingIndex(UniformBuffer.BINDING_MODEL);
        if(!this.transform.init()) {
            Logger.INSTANCE.error("Initialization of gui uniform buffer failed!");
            return false;
        }
        return true;
    }
    
    /**
     * Zeichnet die Komponente.
     */
    public void onRender() {
        if(this.texture == null) {
            this.setNeedsRepaintIf(REPAINT_ON_NULL_TEXTURE);
        }
        if(this.needsRepaint) {
            //Logger.INSTANCE.info("Repainting %s.", this);
            this.repaint();
            this.needsRepaint = false;
        }
        if(this.texture != null) {
            if(this.componentChanged) {
                this.updateBuffer();
                this.componentChanged = false;
            }
            Gui.GUI_EFFECT.use();
            this.transform.bind();
            this.texture.connectToSlot(Gui.EFFECT_TEX_SLOT);
            this.texture.bind();
            //Gui.QUAD.bind();
            Gui.QUAD.draw();
        }
    }
    
    /**
     * Blendet die Komponente langsam aus.
     */
    public void hide() {
        EngineProcess process = new EngineProcess() {
            @Override
            protected void onUpdate(long dMillis) {
                if(Component.this.transparency <= 0.0f) {
                    Component.this.setVisible(false);
                    Component.this.setTransparency(1.0f);
                    this.succeed();
                } else {
                    Component.this.setTransparency(Component.this.transparency - 2e-3f * (float)dMillis);
                }
            }
        };
        this.getGui().getProcessManager().attachProcess(process);
    }
    
    /**
     * Blendet die Komponente langsam ein.
     */
    public void show() {
        if(!this.isVisible()) {
            this.setTransparency(0.0f);
            this.setVisible(true);
            EngineProcess process = new EngineProcess() {
                @Override
                protected void onUpdate(long dMillis) {
                    if(Component.this.transparency >= 1.0f) {
                        Component.this.setTransparency(1.0f);
                        this.succeed();
                    } else {
                        Component.this.setTransparency(Component.this.transparency + 1e-3f * (float)dMillis);
                    }
                }
            };
            this.getGui().getProcessManager().attachProcess(process);
        }
    }
    
    /**
     * Liefert das Fenster, das diese Komponente enthaelt.
     * @return 
     */
    public Window getParentWindow() {
        if(this instanceof Window) {
            return (Window)this;
        } else if(this.container == null) {
            return null;
        } else {
            return this.container.getParentWindow();
        }
    }
    
    /**
     * Liefert eine Komponente anhand ihres Qualifiers. Dabei ist der Qualifier
     * der zusammengesetzte String aus allen Komponenten von der Wurzel bis zum
     * gesuchten Element, getrennt mit Punkten.
     * @param qualifier Der Qualifier
     * @return Die Komponente oder <tt>null</tt>, falls keine Komponente mit
     *         dem uebergebenen Qualifier existiert oder der Qualifier ungueltig
     *         ist
     */
    public Component findChildByQualifier(String qualifier) {
        if(qualifier == null || qualifier.trim().isEmpty()) {
            Logger.INSTANCE.warning("Empty qualifier given to %s.", this);
            return null;
        } else if(!qualifier.contains(".")) {
            return this.getDirectChildByName(qualifier);
        } else {
            String split[] = qualifier.trim().split(".", 2);
            Component child = this.getDirectChildByName(split[0]);
            return child == null ? null : child.findChildByQualifier(split[1]);
        }
    }
    
    /**
     * Liefert ein Kindelement anhand seines Namens.
     * @param name Der Name des Kindes
     * @return Das Kindelement oder <tt>null</tt>, falls keines mit dem 
     *         gegebenen Namen existiert
     */
    private Component getDirectChildByName(String name) {
        for(Component child : this.children) {
            if(child.getName().equals(name)) {
                return child;
            }
        }
        Logger.INSTANCE.warning("Component %s has no child with name %s.", this, name);
        return null;
    }
    
    /**
     * Markiert die Komponente als neu zu zeichnend.
     */
    protected void needsRepaint() {
        this.needsRepaint = true;
    }
    
    @Override
    public String toString() {
        return String.format("%s '%s' (ID: %s)", this.getClass().getSimpleName(), this.getName(), this.getId());
    }
}
