/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package main;

import actor.Actor;
import actor.ActorFactory;
import event.events.ActorCreatedEvent;
import event.events.ActorDeletedEvent;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import logging.Logger;
import org.w3c.dom.Node;
import view.GameView;
import view.HumanView;



/**
 * Basisklasse der GameLogic. Bietet die Moeglichkeit, Actors zu erstellen, zu
 * loeschen und zu verwalten und GameViews zu registrieren und loeschen. In
 * einem eigenen Projekt sollte von dieser Klasse abgleitet werden.
 * @author nico3000
 */
public abstract class BaseGameLogic {
    /**
     * Klasse, um einen Actor erstellen zu koennen.
     */
    private final ActorFactory globalActorFactory = new ActorFactory();
    /**
     * Alle erfolgreich erstellten Instanzen der Klasse Actor nach ID sortiert.
     */
    private final Map<Integer, Actor> actorsById = new HashMap<>();
    /**
     * Alle erfolgreich erstellten Instanzen der Klasse Actor, deren Name nicht
     * <tt>null</tt> ist, nach Name sortiert.
     */
    private final Map<String, Actor> actorsByName = new HashMap<>();
    /**
     * Alle registrierten Views.
     */
    private final List<GameView> gameViews = new LinkedList<>();
    /**
     * Die <tt>HumanView</tt> des Spiels.
     */
    private HumanView humanView = new HumanView();
        
    /**
     * Initialisiert die GameLogic.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        if(!this.globalActorFactory.init()) {
            Logger.INSTANCE.error("ActorFactory could not be initialized.");
            return false;
        }
        if(!this.registerView(this.humanView)) {
            Logger.INSTANCE.error("HumanView could not be registered!");
            return false;
        }
        return true;
    }

    /**
     * Liefert die <tt>HumanView</tt>.
     * @return Die <tt>HumanView</tt>
     */
    public HumanView getHumanView() {
        return this.humanView;
    }

    /**
     * Liefert die <tt>HumanView</tt>.
     * @return Die <tt>HumanView</tt>
     */
    public void setHumanView(HumanView view) {
    	this.humanView = view;
    }
    
    /**
     * Ruft die Methode <tt>update()</tt> aller registrierten Instanzen der
     * Klassen <tt>Actor</tt> und <tt>GameView</tt> auf.
     * @param deltaMillis Dauer des letzten Frames in Millisekunden (Gametimer)
     */
    public void update(long deltaMillis) {
        for(Actor actor : this.actorsById.values()) {
            actor.update(deltaMillis);
        }
        for(GameView view : this.gameViews) {
            view.onUpdate(deltaMillis);
        }
    }
    
    /**
     * Erzeugt einen Actor und fuegt ihn der globalen Menge aller Actorinstanzen
     * hinzu.
     * @param resource Die XML Resource des Actors
     * @param overrideData Eigene Daten, die die Standarddaten ueberschreiben
     *                     sollen (kann <tt>null</tt> sein)
     * @return Der neue Actor oder <tt>null</tt>, falls es einen Fehler gab
     */
    public Actor createActor(String resource, Node overrideData) {
        Actor actor = this.globalActorFactory.createActor(resource, overrideData);
        if(actor != null) {
            this.actorsById.put(actor.getId(), actor);
            if(actor.getName() != null && this.actorsByName.put(actor.getName(), actor) != null) {
                Logger.INSTANCE.warning("two or more actors with same name (" + actor.getName() + ")");
            }
            GameApp.getEventManager().triggerEvent(new ActorCreatedEvent(actor.getId()));
        }
        return actor;
    }
    
    /**
     * Getter fuer einen Actor anhand seiner ID.
     * @param id Die ID des Actors
     * @return Der Actor, oder <tt>null</tt> bei ungueltiger ID
     */
    public Actor getActor(int id) {
        return this.actorsById.get(id);
    }
    
    /**
     * Getter fuer einen Actor anhand seines Namens.
     * @param name Der Name des Actors
     * @return Der Actor, oder <tt>null</tt> bei ungueltigem Namen
     */
    public Actor getActor(String name) {
        return this.actorsByName.get(name);
    }
    
    /**
     * Liefert fuer einen Actor Namen seine zugehoerige ID.
     * @param name Der Name
     * @return Die ID, oder <tt>ActorFactory.INVALID_ACTOR_ID</tt>, wenn kein
     *         Actor mit dem angegeben Namen existiert
     */
    public int getActorId(String name) {
        if(name == null) {
            return ActorFactory.INVALID_ACTOR_ID;
        } else {
            Actor actor = this.actorsByName.get(name);
            return actor == null ? ActorFactory.INVALID_ACTOR_ID : actor.getId();
        }
    }
    
    /**
     * Loescht einen Actor anhand seiner ID.
     * @param id Die ID des zu loeschenden Actors
     */
    public void deleteActor(int id) {
        Actor actor = this.getActor(id);
        if(actor != null) {
            actor.destroy();
            this.actorsById.remove(id);
            if(actor.getName() != null) {
                this.actorsByName.remove(actor.getName());
            }
            GameApp.getEventManager().queueEvent(new ActorDeletedEvent(id));
        }
    }
    
    /**
     * Ruft die Methode <tt>restore()</tt> aller registrierten GameViews auf.
     */
    public void restore() {
        for(GameView view : this.gameViews) {
            view.onRestore();
        }
    }
    
    /**
     * Ruft die Methode <tt>render()</tt> aller registrierten GameViews auf.
     */
    public void render() {
        for(GameView view : this.gameViews) {
            view.onRender();
        }
    }
    
    /**
     * Initialisiert und registriert eine View.
     * @param view Die View
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean registerView(GameView view) {
        if(view.init()) {
            this.gameViews.add(view);
            view.onRestore();
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Entfernt und zerstoert eine View.
     * @param view Die View
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean removeView(GameView view) {
        if(this.gameViews.remove(view)) {
            view.destroy();
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Ruft die Methode <tt>destroy()</tt> aller registrierten Actorinstanzen
     * und GameViews auf.
     */
    public void destroy() {
        for(GameView view : this.gameViews) {
            view.destroy();
        }
        this.gameViews.clear();
        for(Actor actor : this.actorsById.values()) {
            actor.destroy();
        }
        this.actorsById.clear();
        this.actorsByName.clear();
    }
    
    /**
     * Standardmethode, um einen Level zu laden.
     * @param resource De Resource mit den Leveldaten
     */
    public abstract void loadLevel(String resource);
}
