/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package main;

import event.EventData;
import event.EventListener;
import event.EventManager;
import event.events.QuitEvent;
import graphics.GraphicsLayer;
import graphics.geometry.Geometry;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;
import gui.Gui;
import input.InputHandler;
import input.PanicKeyboardHandler;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.logging.Level;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import logging.Logger;
import mesh.Mesh;
import mesh.MeshLoader;
import mesh.material.MaterialLibrary;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import process.ProcessManager;
import resource.ResHandle;
import resource.Resource;
import resource.ThreadedResCache;
import resource.file.DirectoryResourceFile;
import scripting.JavaScriptManager;
import scripting.ScriptManager;
import config.Config;

/**
 *
 * @author nico3000
 */
public class GameApp {
    private static boolean bContinue = true;
    private static boolean debugMode = false;
    private static Config config = null;
    private static final ThreadedResCache resCache = new ThreadedResCache(100);
    private static final ScriptManager scriptManager = new JavaScriptManager();
    private static final Timer gameTimer = new Timer(Timer.Precision.MILLISCONDS);
    private static final Timer realTimer = new Timer(Timer.Precision.MILLISCONDS);
    private static final InputHandler inputHandler = new InputHandler();
    private static final GraphicsLayer graphicsLayer = new GraphicsLayer();
    private static final PhysicsLayer physicsLayer = new PhysicsLayer();
    private static final EventManager eventManager = new EventManager();
    private static BaseGameLogic gameLogic = null;
    /**
     * Globale Prozessverwaltung.
     */
    private static final ProcessManager globalProcessManager = new ProcessManager();
    /**
     * Builder fuer XML Resourcen.
     */
    private static DocumentBuilder documentBuilder = null;
    
    public static void setGameLogic(BaseGameLogic gameLogic) {
        GameApp.gameLogic = gameLogic;
    }
    
    public static void setConfigFile(InputStream defaultConfigFile) {
        config = new Config("config.ini", defaultConfigFile);
    }
            
    public static boolean init() {
        if(gameLogic == null) {
            System.err.println("No GameLogic set!");
            return false;
        }
        if(config == null) {
            System.err.println("No config file set!");
            return false;
        }
        if(!Logger.INSTANCE.open("engine.log")) {
            System.err.println("Logger could not be open file engine.log!");
        }
        if(!config.init()) {
            Logger.INSTANCE.error("Configuration file could not be initialized!");
            return false;
        }
        debugMode = config.getAsBoolean("bDebugMode");
        resCache.setFile(new DirectoryResourceFile(config.getAsString("sResourcePath")));
        if(!resCache.init()) {
            Logger.INSTANCE.error("Resource cache could not be initialized!");
            return false;
        }
        if(!scriptManager.init()) {
            Logger.INSTANCE.error("ScriptManager could not be initialized!");
            return false;
        }
        DocumentBuilderFactory fac = DocumentBuilderFactory.newInstance();
        try {
            documentBuilder = fac.newDocumentBuilder();
        } catch (ParserConfigurationException ex) {
            Logger.INSTANCE.error(ex.getMessage());
            return false;
        }
        gameTimer.init();
        if(!graphicsLayer.init()) {
            Logger.INSTANCE.error("GraphicsLayer could not be initialized!");
            return false;
        }
        graphicsLayer.setClearColor(0.05f, 0.07f, 0.1f, 1.0f);
        if(!inputHandler.init()) {
            Logger.INSTANCE.error("Input could not be initialized!");
            return false;            
        }
        physicsLayer.init();
        if(!physicsLayer.init()) {
            Logger.INSTANCE.error("PhysicsLayer could not be initialized!");
            return false;
        }
        inputHandler.registerKeyboardHandler(new PanicKeyboardHandler());
        eventManager.register(new EventListener() {
            @Override
            public void trigger(EventData data) {
                GameApp.bContinue = false;
            }
        }, QuitEvent.ID);
        if(!gameLogic.init()) {
            Logger.INSTANCE.error("GameLogic could not be initialized!");
            return false;   
        }
        if(!TextureLoader.init()) {
            Logger.INSTANCE.error("TextureLoader could not be initialized!");
            return false;
        }
        if(!MeshLoader.init()) {
            Logger.INSTANCE.error("MeshLoader could not be initialized!");
            return false;
        }
        if(!Gui.init()) {
            Logger.INSTANCE.error("Gui could not be initialized!");
            return false;
        }
        
        //FreeCamera freeCamera = new FreeCamera();
        //inputHandler.registerKeyboardHandler(freeCamera);
        //inputHandler.registerPointerHandler(freeCamera);
        //gameLogic.registerView(freeCamera);
        
        return true;
    }

    public static Config getConfig() {
        return config;
    }

    public static Timer getGameTimer() {
        return gameTimer;
    }

    public static Timer getRealTimer() {
        return realTimer;
    }

    public static InputHandler getInputHandler() {
        return inputHandler;
    }

    public static EventManager getEventManager() {
        return eventManager;
    }

    public static BaseGameLogic getGameLogic() {
        return gameLogic;
    }

    public static PhysicsLayer getPhysicsLayer() {
        return physicsLayer;
    }

    public static GraphicsLayer getGraphicsLayer() {
        return graphicsLayer;
    }

    public static ThreadedResCache getResCache() {
        return resCache;
    }

    public static ScriptManager getScriptManager() {
        return scriptManager;
    }

    public static ProcessManager getGlobalProcessManager() {
        return globalProcessManager;
    }
    
    public static boolean isInDebugMode() {
        return debugMode;
    }
    
    public static void mainloop() {
        Logger.INSTANCE.info("---- Entering mainloop ----");
        while(bContinue && !graphicsLayer.isWindowCloseRequested()) {
            //Logger.INSTANCE.info("---- Next Frame ----");
            graphicsLayer.clear(GraphicsLayer.MASK_COLOR | GraphicsLayer.MASK_DEPTH | GraphicsLayer.MASK_STENCIL);
            gameTimer.next();
            realTimer.next();
            inputHandler.processInput();
            eventManager.processEvents();
            physicsLayer.update(gameTimer.getDelta());
            globalProcessManager.update(gameTimer.getDelta());
            gameLogic.update(gameTimer.getDelta());
            gameLogic.render();
            graphicsLayer.update(realTimer.getDelta());
            graphicsLayer.setWindowTitle("FPS: " + graphicsLayer.getFps());
        }
        Logger.INSTANCE.info("---- Leaving mainloop ----");
    }
    
    public static void destroy() {
        Gui.destroy();
        MeshLoader.destroy();
        TextureLoader.destroy();
        gameLogic.destroy();
        inputHandler.destroy();
        graphicsLayer.destroy();
        physicsLayer.destroy();
        scriptManager.destroy();
        resCache.destroy();
        if(isInDebugMode()) {
            String status = "Engine debugging information\n";
            status += "++ begin ++\n";
            status += Mesh.getDebugStatus() + "\n";
            status += Geometry.getDebugStatus() + "\n";
            status += MaterialLibrary.getDebugStatus() + "\n";
            status += Texture.getDebugStatus() + "\n";
            status += "++  end  ++";
            Logger.INSTANCE.info(status);
        }
        config.save();
        Logger.INSTANCE.close();
    }
    
    public static Document loadXML(String resource) {
        ResHandle handle = resCache.getHandle(new Resource(resource));
        if(handle == null) {
            Logger.INSTANCE.error("failed to open resource " + resource);
            return null;
        }
        InputStream stream = handle.getBufferAsInputStream();
        try {
            return documentBuilder.parse(stream, null);
        } catch (SAXException | IOException ex) {
            Logger.INSTANCE.error(ex.getMessage());
            return null;
        } finally {
            try {
                stream.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            handle.unlock();
        }
    }
    
    public static Node loadXMLNodeFromString(String xmlString) {
    	InputSource stream = new InputSource(new StringReader(xmlString));
    	try {
    		return documentBuilder.parse(stream).getFirstChild();
    	} catch (SAXException | IOException ex) {
    		Logger.INSTANCE.error(ex.getMessage());
    		return null;
    	}
    }
    
    public static String getSource(String file) {
        BufferedReader reader = null;
        try {
            String src = "";
            reader = new BufferedReader(new FileReader(file));
            String line;
            while((line = reader.readLine()) != null) {
                src += line + "\n";
            }
            return src;
        } catch (IOException ex) {
            java.util.logging.Logger.getLogger(GameApp.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            try {
                if(reader != null) {
                    reader.close();
                }
            } catch (IOException ex) {
                java.util.logging.Logger.getLogger(GameApp.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return null;
    }
}
