/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package main;

/**
 * Timerklasse mit einer Afloesung von 10^-9s (Nanosekunden). Benutzt zur
 * Messung die Methode <tt>System.nanoTime()</tt>.
 * @author nico3000
 */
public class Timer {
    /**
     * Moegliche Praezision eines Timers.
     */
    public static enum Precision {
        /**
         * Praezision Nanosekunden (10^-9s)
         */
        NANOSECONDS,
        /**
         * Praezision Mikrosekunden (10^-6s)
         */
        MICROSECONDS,
        /**
         * Praezision Millisekunden (10^-3s)
         */
        MILLISCONDS,
        /**
         * Praezision Sekunden (s)
         */
        SECONDS,
    }
    
    /**
     * Letzter Rueckgabewert von <tt>System.nanoTime()</tt>.
     */
    private long lastStop;
    
    /**
     * Zeiteinheiten zwischen den letzten zwei Aufrufen von <tt>next()</tt>.
     */
    private long delta;
    
    /**
     * Divisor fuer die Anzahl Zeiteinheiten, die von <tt>getDeltaUnits()</tt>
     * zurueckgegeben wird.
     */
    private long globalDivisor;
    
    /**
     * Interner Akkumulator fuer die korrekte Zeitrechnung bei Divisor != 1.
     */
    private long accumulator;
    
    /**
     * Interner Zaehler des Zeitfaktors.
     */
    private long numerator = 1L;
    
    /**
     * Interner Nenner des Zeitfaktors.
     */
    private long denominator = 1L;
    /**
     * Aktueller Zeitstempel, startet bei 0.
     */
    private long timeStamp = 0;

    /**
     * Konstruktor fuer einen Timer.
     * @param precision Einheit des Rueckgabewertes von <tt>getDelta()</tt>
     */
    public Timer(Precision precision) {
        this.globalDivisor = 1L;
        switch(precision) {
            case SECONDS:      this.globalDivisor *= 1000L;
            case MILLISCONDS:  this.globalDivisor *= 1000L;
            case MICROSECONDS: this.globalDivisor *= 1000L;
        }
    }
    
    /**
     * Initialisiert den Timer. Sollte genau einnmal aufgerufen werden.
     */
    public void init() {
        this.lastStop = System.nanoTime();
        this.delta = 0L;
        this.accumulator = 0L;
    }
    
    /**
     * Ermittelt die Zweit seit dem letzten Aufruf von <tt>next()</tt> (oder
     * <tt>init()</tt>).
     */
    public void next() {
        long now = System.nanoTime();
        long realdelta = now - this.lastStop + this.accumulator;
        this.delta = (realdelta * this.denominator / (this.numerator * this.globalDivisor));
        this.accumulator = realdelta - this.delta * (this.numerator * this.globalDivisor) / this.denominator;
        this.lastStop = now;
        this.timeStamp += this.delta;
    }
    
    /**
     * Liefert die Anzahl der Zeiteinheiten zwischen den letzten beiden Aufrufen
     * von <tt>next()</tt>.
     * @return Vergangene Zeit in der Einheit, die dem Konstruktor uebergeben
     *         wurde
     */
    public long getDelta() {
        return this.delta;
    }

    /**
     * Liefert den aktuellen Zeitstempel, mit dem ein Frame eindeutig
     * identifiziert werden kann.
     * @return Der aktuelle Zeitstempel
     */
    public long getTimeStamp() {
        return timeStamp;
    }
    
    /**
     * Setzt den Zeitfaktor <tt>f = numerator / denominator</tt> des Timers. Ist
     * <tt>f &lt; 1</tt>, vergeht die Timerzeit langsamer als die Echtzeit,
     * andernfalls schneller.
     * @param numerator Zaehler des Faktors
     * @param denominator Nenner des Faktors
     */
    public void setFactor(long numerator, long denominator) {
        this.numerator = numerator;
        this.denominator = denominator;
    }
}
