/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh;

import graphics.geometry.Geometry;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import logging.Logger;
import mesh.loader.GeometryLoader;
import mesh.material.Material;
import mesh.material.MaterialLibrary;

/**
 * Ein Mesh ist ein Objekt im Raum. Es besteht aus einer Geometrie und einer
 * Menge von Materialien.
 * @author nico3000
 */
public class Mesh {
    /**
     * Alle erzeugten Meshes.
     */
    private static final List<Mesh> INSTANCES = new LinkedList<>();
    
    /**
     * Erzeugt ein neues Mesh mithilfe eines GeometryLoaders. Der Loader darf
     * nicht <tt>null</tt> sein.
     * @param loader Der Loader, der fuer das Laden der Geometrie verantwortlich
     *               ist
     * @return Das neue Mesh oder <tt>null</tt>, falls kein Loader angegeben
     *         wurde
     */
    public static Mesh createMesh(GeometryLoader loader) {
        if(loader == null) {
            return null;
        }
        Geometry geo = loader.load();
        if(geo != null) {
            Mesh mesh = new Mesh(geo);
            MaterialLibrary customLibrary = null;
            if(loader.getCustomMaterialLibrary() != null) {
                customLibrary = MaterialLibrary.getLibrary(loader.getCustomMaterialLibrary());
            }
            boolean customUsed = false;
            for(Map.Entry<Material, IndexBufferInterval> entry : geo.getIntervals()) {
                Material customMaterial = customLibrary == null ? null : customLibrary.getMaterialByName(entry.getKey().getName());
                if(customMaterial == null) {
                    mesh.intervals.put(entry.getKey(), entry.getValue());
                    entry.getKey().addRef();
                    MaterialLibrary library = MaterialLibrary.getLibrary(entry.getKey().getLibraryName());
                    if(!mesh.referencedLibraries.add(library)) {
                        library.release();
                    }
                } else {
                    customUsed = true;
                    mesh.intervals.put(customMaterial, entry.getValue());
                    mesh.referencedLibraries.add(customLibrary);
                }
            }
            if(!customUsed) {
                if(customLibrary != null) {
                    Logger.INSTANCE.warning("Referenced custom MaterialLibrary %s not used.", customLibrary.getName());
                    customLibrary.release();
                }
            } else {
                mesh.referencedLibraries.add(customLibrary);
            }
            loader.release();
            INSTANCES.add(mesh);
            mesh.addRef();
            return mesh;
        } else {
            return null;
        }
    }
    
    /**
     * Gibt den Status der bekannten Meshes zurueck.
     * @return Der Status
     */
    public static String getDebugStatus() {
        String meshes = "";
        for(Mesh mesh : INSTANCES) {
            if(mesh.geometry != null) {
                meshes += "\n- " + mesh.geometry.getResource();
            }
        }
        return String.format("Active Mesh instances: %s", meshes.isEmpty() ? "none" : meshes);
    }
    
    /**
     * Die Intervalle des Indexbuffers gehoeren zu bestimmten Materialien.
     */
    private Map<Material, IndexBufferInterval> intervals = new HashMap<>();
    /**
     * Die Geometriedaten des Meshes.
     */
    private Geometry geometry;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * Alle referenzierten MaterialLibrariy Instanzen.
     */
    private final Set<MaterialLibrary> referencedLibraries = new HashSet<>();
    /**
     * Solange das Mesh nicht geladen ist, der Zeitstempel des Erstellens, sonst
     * die Ladezeit in Millisekunden.
     */
    private long loadingTime = System.currentTimeMillis();
    /**
     * Die Anzahl der Vertices.
     */
    private int vertexCount = 0;
    /**
     * Die Anzahl der Dreiecke.
     */
    private int triangleCount = 0;
    /**
     * Groesse des Meshes im VRAM.
     */
    private int size = 0;

    /**
     * Konstruktor.
     * @param geometry Die Geometriedaten des Meshes
     */
    private Mesh(Geometry geometry) {
        this.geometry = geometry;
    }
    
    /**
     * Erhoeht den Reference Counter um 1.
     */
    void addRef() {
        ++this.ref;
    }
    
    /**
     * Zeichnet das Mesh. Dazu wird zunaechst die Gemetrie gebunden und danach
     * werden die einzelnen Bereiche mit ihrem jeweiligen Material nacheinander
     * gezeichnet.
     */
    public void draw() {
        if(this.geometry != null) {
            this.geometry.bind();
            if(this.intervals.isEmpty()) {
                this.geometry.render();
            } else {
                for(Material material : this.intervals.keySet()) {
                    material.bind();
                    IndexBufferInterval interval = this.intervals.get(material);
                    this.geometry.drawInterval(interval.getStart(), interval.getCount());
                }
            } 
        } else {
            Logger.INSTANCE.warning("Drawing empty Mesh skipped.");
        }        
    }
    
    /**
     * Verringert den Reference Counter um 1 und loescht die Resource, falls er
     * kleiner oder gleich 0 wird.
     */
    public void release() {
        if(--this.ref == 0) {
            for(Material material : this.intervals.keySet()) {
                if(material != null) {
                    material.release();
                }
            }
            this.intervals.clear();
            for(MaterialLibrary library : this.referencedLibraries) {
                library.release();
            }
            this.referencedLibraries.clear();
            this.geometry.release();
            this.geometry = null;
        }
    }
    
    /**
     * Liefert einen String mit Statistiken des Meshes.
     * @return Statistiken
     */
    public String getStats() {
        return String.format("%s stats: %d vertices; %d triangles; %d materials; %.2fs loading time",
                             this.geometry.getResource(), this.vertexCount, this.triangleCount,
                             this.intervals.size(), 1e-3f * (float)(this.loadingTime));
    }
}
