/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh.material;

import graphics.shader.Effect;
import graphics.shader.UniformBuffer;
import graphics.texture.SlotAdmin;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;

import java.nio.FloatBuffer;

import logging.Logger;

import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Vector4f;

/**
 * Ein Material eines Meshes.
 * @author nico3000
 */
public class Material {
    public static final int AMBIENT_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int DIFFUSE_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int SPECULAR_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int ALPHA_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    /**
     * Der Name des Materials.
     */
    private final String name;
    /**
     * Der Name der MaterialLibrary, der dieses Material untergeordnet ist.
     */
    private final String libraryName;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * UniformBuffer der Daten im Shader.
     */
    private final UniformBuffer buffer = new UniformBuffer();
    /**
     * Die ambiente Farbe.
     */
    private final Vector4f ka = new Vector4f();
    /**
     * Die diffuse Farbe.
     */
    private final Vector4f kd = new Vector4f();
    /**
     * Die spekulare Farbe.
     */
    private final Vector4f ks = new Vector4f();
    /**
     * Textur fuer den Alphakanal.
     */
    private Texture mapAlpha = null;
    /**
     * Textur fuer den ambienten Anteil.
     */
    private Texture mapAmbient = null;
    /**
     * Textur fuer den diffusen Anteil.
     */
    private Texture mapDiffuse = null;
    /**
     * Textur fuer den spekularen Anteil.
     */
    private Texture mapSpecular = null;
    
    private String resourceAlpha = null;
    private String resourceAmbient = null;
    private String resourceDiffuse = null;
    private String resourceSpecular = null;
    private int size = 0;

    /**
     * Konstruktor.
     * @param libraryName Der Name der MaterialLibrary, der dieses Material
     *                    untergeordnet ist
     * @param name Der Name des Materials
     */
    Material(String libraryName, String name) {
        this.libraryName = libraryName;
        this.name = name;
    }

    /**
     * Liefert den Namen des Materials.
     * @return Der Name
     */
    public String getName() {
        return this.name;
    }

    /**
     * Liefert den Namen der diesem Material zugeordneten MaterialLibrary.
     * @return Der Name der MaterialLibrary
     */
    public String getLibraryName() {
        return this.libraryName;
    }
    
    /**
     * Initialisiert den Buffer des Materials.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        if(!this.buffer.init()) {
            Logger.INSTANCE.error("Buffer of Material %s could not be initialized!", this.name);
            return false;
        }
        this.buffer.setBindingIndex(UniformBuffer.BINDING_MATERIAL);
        this.size = 16 * 4 + (this.mapAlpha != null ? this.mapAlpha.getSize() : 0)
                           + (this.mapAmbient != null ? this.mapAmbient.getSize() : 0)
                           + (this.mapDiffuse != null ? this.mapDiffuse.getSize() : 0)
                           + (this.mapSpecular != null ? this.mapSpecular.getSize() : 0);
        return true;
    }
    
    /**
     * Bindet das Material an den Uniformblock <tt>MaterialBlock</tt>.
     */
    public void bind() {
        if(this.mapAmbient != null) {
            this.mapAmbient.connectToSlot(AMBIENT_TEX_SLOT);
        }
        if(this.mapDiffuse != null) {
            this.mapDiffuse.connectToSlot(DIFFUSE_TEX_SLOT);
        }
        if(this.mapSpecular != null) {
            this.mapSpecular.connectToSlot(SPECULAR_TEX_SLOT);
        }
        if(this.mapAlpha != null) {
            this.mapAlpha.connectToSlot(ALPHA_TEX_SLOT);
        }
        this.buffer.bind();
    }
    
    /**
     * Aktualisiert den das Material. Muss aufgerufen werden, wenn mittels
     * <tt>getK*()</tt> Aenderungen an den Farben oder <tt>set*Texture()</tt>
     * Aenderungen an den Texturen vorgenommen wurden.
     * @return <tt>true</tt>, falls alle Texturen erfolgreich geladen wurden
     */
    public boolean updateData() {
        boolean success = this.updateTextures();
        FloatBuffer data = BufferUtils.createFloatBuffer(16);
        this.ka.store(data);
        this.kd.store(data);
        this.ks.store(data);
        data.put(this.mapAmbient == null ? 0.0f : 1.0f);
        data.put(this.mapDiffuse == null ? 0.0f : 1.0f);
        data.put(this.mapSpecular == null ? 0.0f : 1.0f);
        data.put(this.mapAlpha == null ? 0.0f : 1.0f);
        data.position(0);
        this.buffer.setData(data);
        return success;
    }
    
    /**
     * Aktualisiert die Texturen des Materials.
     * @return <tt>true</tt>, falls alle Texturen erfolgreich geladen wurden
     */
    private boolean updateTextures() {
        if(this.mapAlpha != null && !this.mapAlpha.getResource().equals(this.resourceAlpha)) {
            this.mapAlpha.release();
            this.mapAlpha = null;
        }
        if(this.mapDiffuse != null && !this.mapDiffuse.getResource().equals(this.resourceDiffuse)) {
            this.mapDiffuse.release();
            this.mapDiffuse = null;
        }
        if(this.mapAmbient != null && !this.mapAmbient.getResource().equals(this.resourceAmbient)) {
            this.mapAmbient.release();
            this.mapAmbient = null;
        }
        if(this.mapSpecular != null && !this.mapSpecular.getResource().equals(this.resourceSpecular)) {
            this.mapSpecular.release();
            this.mapSpecular = null;
        }
        boolean success = true;
        if(this.resourceAlpha != null && (this.mapAlpha = TextureLoader.getTexture(this.resourceAlpha)) == null) {
            Logger.INSTANCE.warning("Could not load alpha texture %s of Material %s.", this.resourceAlpha, this.name);
            success = false;
        }
        if(this.resourceAmbient != null && (this.mapAmbient = TextureLoader.getTexture(this.resourceAmbient)) == null) {
            Logger.INSTANCE.warning("Could not load ambient texture %s of Material %s.", this.resourceAmbient, this.name);
            success = false;
        }
        if(this.resourceDiffuse != null && (this.mapDiffuse = TextureLoader.getTexture(this.resourceDiffuse)) == null) {
            Logger.INSTANCE.warning("Could not load diffuse texture %s of Material %s.", this.resourceDiffuse, this.name);
            success = false;
        }
        if(this.resourceSpecular != null && (this.mapSpecular = TextureLoader.getTexture(this.resourceSpecular)) == null) {
            Logger.INSTANCE.warning("Could not load specular texture %s of Material %s.", this.resourceSpecular, this.name);
            success = false;
        }
        return success;
    }
    
    /**
     * Bindet die Variablen eines Effekts an die entsprechenden Slots fuer die
     * Texturen.
     * @param effect Der Effekt
     */
    public static void modifyEffectToUseMaterial(Effect effect) {
        effect.use();
        effect.setUniform("g_AmbientTex", Material.AMBIENT_TEX_SLOT);
        effect.setUniform("g_DiffuseTex", Material.DIFFUSE_TEX_SLOT);
        effect.setUniform("g_SpecularTex", Material.SPECULAR_TEX_SLOT);
        effect.setUniform("g_AlphaTex", Material.ALPHA_TEX_SLOT);
    }
    
    /**
     * Setzt eine Resource fuer die ambiente Textur.
     * @param resource Der Name der Resource
     */
    public void setAmbientTextue(String resource) {
        this.resourceAmbient = resource;
    }
    
    /**
     * Setzt eine Resource fuer die diffuse Textur.
     * @param resource Der Name der Resource
     */
    public void setDiffuseTextue(String resource) {
        this.resourceDiffuse = resource;
    }
    
    /**
     * Setzt eine Resource fuer die spekulare Textur.
     * @param resource Der Name der Resource
     */
    public void setSpecularTextue(String resource) {
        this.resourceSpecular = resource;
    }
    
    /**
     * Setzt eine Resource fuer die Alphakanaltextur.
     * @param resource Der Name der Resource
     */
    public void setAlphaTextue(String resource) {
        this.resourceAlpha = resource;
    }

    /**
     * Liefert die ambiente Farbe.
     * @return Die ambiente Farbe
     */
    public Vector4f getKa() {
        return ka;
    }

    /**
     * Liefert die diffuse Farbe.
     * @return Die diffuse Farbe
     */
    public Vector4f getKd() {
        return kd;
    }

    /**
     * Liefert die spekulare Farbe.
     * @return Die spekulare Farbe
     */
    public Vector4f getKs() {
        return ks;
    }

    public int getSize() {
        return size;
    }
    
    /**
     * Erhoeht den Reference Counter um 1.
     */
    public void addRef() {
        ++this.ref;
    }
    
    /**
     * Verringert den Reference Counter um 1. Ist er danach kleiner oder gleich
     * 0, wird das Material zerstoert.
     */
    public void release() {
        if(this.ref == 0) {
            Logger.INSTANCE.warning("Releasing Material %s impossible because reference counter is already 0!", this.name);
        } else {
            --this.ref;
        }
    }
    
    /**
     * Liefert den aktuellen Reference Count.
     * @return Der aktuelle Reference Count
     */
    int getReferenceCount() {
        return this.ref;
    }
    
    /**
     * Zerstoert das Material explizit. Sollte nur von der zugehoerigen
     * MaterialLibrary aufgerufen werden.
     */
    void destroy() {
        this.buffer.destroy();
        if(this.mapAlpha != null) {
            this.mapAlpha.release();
            this.mapAlpha = null;
        }
        if(this.mapAmbient != null) {
            this.mapAmbient.release();
            this.mapAmbient = null;
        }            
        if(this.mapDiffuse != null) {
            this.mapDiffuse.release();
            this.mapDiffuse = null;
        }
        if(this.mapSpecular != null) {
            this.mapSpecular.release();
            this.mapSpecular = null;
        } 
    }
    
    @Override
    public String toString() {
    	return String.format("Material \"%s\"", this.name);
    }
}
