/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package view;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import logging.Logger;
import main.GameApp;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import util.XMLUtil;
import view.scene.SceneScreenElement;
import actor.Actor;
import actor.ActorFactory;
import actor.components.CameraComponent;
import event.EventData;
import event.EventListener;
import event.events.ActorCreatedEvent;
import event.events.ActorDeletedEvent;
import event.events.ActorEvent;
import event.events.PlayerActorChangedEvent;
import gui.ComponentDrawer;
import gui.DefaultComponentDrawer;
import gui.Gui;
import gui.GuiAction;
import gui.components.Button;
import gui.components.Component;
import gui.components.Container;
import gui.components.Label;
import gui.components.TextArea;
import gui.components.TextField;
import gui.components.ToggleButton;
import gui.components.Window;

/**
 * Die HumanView verwaltet die verfuegbaren Szenen, kuemmert sich um den Input
 * des Spielers und die Audio-/Videoausgabe.
 * @author nico3000
 */
public class HumanView extends GameView implements EventListener {
    /**
     * Die grafischen Elemente der <tt>HumanView</tt>.
     */
    private final List<ScreenElement> screenElements = new LinkedList<>();
    /**
     * Alle bekannten Guis.
     */
    private final Map<String, Gui> guis = new HashMap<>();
    /**
     * Aktuelle Szene.
     */
    private SceneScreenElement scene;
    /**
     * Die aktuelle Kamera.
     */
    private Actor playerActor = null;
    /**
     * Verwaltet die Events, die bestimmten Inputs zugeordnet werden.
     */
    private final PlayerControls playerControls = new PlayerControls();
    
    private final PlayerControlsValidator validator = new PlayerControlsValidator();   
    /**
     * Die Drawerklasse fuer Gui Komponenten.
     */
    private ComponentDrawer drawer = new DefaultComponentDrawer();

    @Override
    public boolean init() {
        if(!this.playerControls.init()) {
            Logger.INSTANCE.error("PlayerControls could not be initialized!");
            this.destroy();
            return false;
        }
        if(GameApp.getConfig().getAsBoolean("bDefaultFpsControls") && !this.validator.init()) {
            Logger.INSTANCE.error("PlayerControlsValidator could not be initialized!");
            this.destroy();
            return false;
        }
        if(!this.drawer.init()) {
            Logger.INSTANCE.error("ComponentDrawer could not be initialized!");
            this.destroy();
            return false;
        }
        GameApp.getInputHandler().registerKeyboardHandler(this.playerControls);
        GameApp.getInputHandler().registerPointerHandler(this.playerControls);
        GameApp.getEventManager().register(this, ActorCreatedEvent.ID, ActorDeletedEvent.ID);
        return true;
    }

    @Override
    public void destroy() {
        for(ScreenElement elem : this.screenElements) {
            elem.onDestroy();
        }        
        GameApp.getEventManager().unregister(this, ActorCreatedEvent.ID, ActorDeletedEvent.ID);
        GameApp.getInputHandler().unregisterKeyboardHandler(this.playerControls);
        GameApp.getInputHandler().unregisterPointerHandler(this.playerControls);
        if(GameApp.getConfig().getAsBoolean("bDefaultFpsControls")) {
            this.validator.destroy();
        }
    }
    
    /**
     * Erzeugt eine neue Gui.
     * @param resource Die Resource
     * @return Die Gui oder <tt>null</tt>, falls ein Fehler aufgetreten ist
     */
    public Gui createGui(String resource) {
        Document doc = GameApp.loadXML(resource);
        Node root = doc.getFirstChild();
        if(root == null || !root.getNodeName().equals("Gui")) {
            Logger.INSTANCE.warning("Xml from resource %s does not start with <Gui>-element.", resource);
            return null;
        }
        String name = XMLUtil.getAttributeAsString(root, "name");
        if(name == null) {
            Logger.INSTANCE.warning("Gui from resource %s has no name.", resource);
            return null;
        }
        if(this.guis.containsKey(name)) {
            Logger.INSTANCE.warning("Two guis with the same name \"%s\" are not allowed!", name);
            return null;
        }
        Gui gui = new Gui(name, this.drawer);
        if(this.pushElement(gui)) {
        	this.guis.put(name, gui);
        } else {
        	return null;
        }
        Node node = root.getFirstChild();
        while(node != null) {
            this.createComponentsRecursive(node, gui, null);
            node = node.getNextSibling();
        }
        return gui;
    }
    
    /**
     * Erzeugt aus einer Node und rekursiv aus allen Kindern Komponenten fuer
     * eine gegebene Gui.
     * @param node Die Node
     * @param gui Die Gui
     * @param parent Die Elternkomponente
     */
    public Component createComponentsRecursive(Node node, Gui gui, Component parent) {
        Component component = createComponent(node);
        
        if(component != null) {
        	component.setId(XMLUtil.getAttributeAsString(node, "id"));
        	gui.addComponent(component);
        	if(parent != null) {
        		parent.addChild(component);
        	}
        	if(!this.setComponentDetails(component, node)) {
        		gui.removeComponent(component);
        	}
        	NodeList list = node.getChildNodes();
        	for(int i=0; i < list.getLength(); ++i) {
        		this.createComponentsRecursive(list.item(i), gui, component);
        	}
        }
        
        return component;
    }

	public Component createComponent(Node node) {
		Component component = null;
        switch(node.getNodeName()) {
            case "Label":
                component = new Label(XMLUtil.getAttributeAsString(node, "text"), XMLUtil.getAttributeAsBoolean(node, "decorated", true));
                break;
            case "Button":
                component = new Button(XMLUtil.getAttributeAsString(node, "text"));
                break;
            case "ToggleButton":
                component = new ToggleButton(XMLUtil.getAttributeAsString(node, "text"));
                break;
            case "TextField":
                component = new TextField(XMLUtil.getAttributeAsString(node, "text"));
                ((TextField)component).setFontSize(XMLUtil.getAttributeAsInt(node, "fontsize", 12));
                break;
            case "Container":
                component = new Container(XMLUtil.getAttributeAsBoolean(node, "transparent", true));
                break;
            case "Window":
                component = new Window(XMLUtil.getAttributeAsString(node, "title"), XMLUtil.getAttributeAsBoolean(node, "closeable", true));
                break;
            case "TextArea":
                component = new TextArea(XMLUtil.getAttributeAsString(node, "text"));
                break;
            case "#comment": // fall-through
            case "#text":
                // nothing to do
                break;
            default:
//                Logger.INSTANCE.error("%s is no known gui component!", node.getNodeName());
                break;
        }
		return component;
	}
    
    /**
     * Hilfsmethode, die die Groesse, Position und Tooltip aus einem XML Node
     * einer Komponente zuweist.
     * @param comp Die Komponente
     * @param node Der Node
     */
    private boolean setComponentDetails(Component comp, Node node) {
        String name = XMLUtil.getAttributeAsString(node, "name");
        if(name != null) {
            comp.setName(name);
        }
        String widthString = XMLUtil.getAttributeAsString(node, "width");
        String heightString = XMLUtil.getAttributeAsString(node, "height");
        if(widthString == null || heightString == null) {
            Logger.INSTANCE.warning("No width / height for component %s.", comp.getQualifier());
            return false;
        }
        int width, height;
        widthString = widthString.trim();
        heightString = heightString.trim();
        try {
            if (widthString.endsWith("%")) {
                float percentage = Float.parseFloat(widthString.substring(0, widthString.length() - 1));
                width = comp.relToAbsWidth(1e-2f * percentage);
            } else {
                width = Integer.parseInt(widthString);
            }
        } catch (NumberFormatException ex) {
            Logger.INSTANCE.warning("Invalid width for component %s.", comp.getQualifier());
            return false;
        }
        try {
            if (heightString.endsWith("%")) {
                float percentage = Float.parseFloat(heightString.substring(0, heightString.length() - 1));
                height = comp.relToAbsHeight(1e-2f * percentage);
            } else {
                height = Integer.parseInt(heightString);
            }
        } catch (NumberFormatException ex) {
            Logger.INSTANCE.warning("Invalid height for component %s.", comp.getQualifier());
            return false;
        }
        comp.setSize(width, height);
        String x = XMLUtil.getAttributeAsString(node, "x");
        if(x != null) {
            try {
                if (x.endsWith("%")) {
                    float percentageX = 1e-2f * Float.parseFloat(x.substring(0, x.length() - 1));
                    comp.setPosition(percentageX, Component.POS_NO_CHANGE_FLOAT);
                } else {
                    int posX = Integer.parseInt(x);
                    comp.setPosition(posX, Component.POS_NO_CHANGE_INT);
                }
            } catch (NumberFormatException ex) {
                Logger.INSTANCE.warning("Invalid x attribute (%s) for component %s.", x, name);
                return false;
            }
        }
        String y = XMLUtil.getAttributeAsString(node, "y");
        if(y != null) {
            try {
                if (y.endsWith("%")) {
                    float percentageY = 1e-2f * Float.parseFloat(y.substring(0, y.length() - 1));
                    comp.setPosition(Component.POS_NO_CHANGE_FLOAT, percentageY);
                } else {
                    int posY = Integer.parseInt(y);
                    comp.setPosition(Component.POS_NO_CHANGE_INT, posY);
                }
            } catch (NumberFormatException ex) {
                Logger.INSTANCE.warning("Invalid y attribute (%s) for component %s.", y, name);
                return false;
            }
        }
        comp.setLocalZOrder(XMLUtil.getAttributeAsInt(node, "z", 0));
        comp.setTooltip(XMLUtil.getAttributeAsString(node, "tooltip"));
        comp.setVisible(XMLUtil.getAttributeAsBoolean(node, "visible", true));
        comp.setAction(GuiAction.fromString(XMLUtil.getAttributeAsString(node, "action")));
        return true;
    }
    
    /**
     * Liefert eine Komponente anhand ihres Qualifiers. Dabei ist der Qualifier
     * der zusammengesetzte String aus dem Namen der Gui und allen Komponenten
     * von der Wurzel bis zum gesuchten Element, getrennt mit Punkten.
     * Alternativ kann auch nach der ID der Komponente mittels
     * <tt>&lt;Name der Gui&gt;#&lt;ID der Komponente&gt;</tt> gesucht werden.
     * @param qualifier Der Qualifier
     * @return Die Komponente oder <tt>null</tt>, falls keine Komponente mit
     *         dem uebergebenen Qualifier existiert oder der Qualifier ungueltig
     *         ist
     */
    public Component getGuiComponent(String qualifier) {
        String split[] = qualifier.split("[.#]", 2);
        if(split.length != 2) {
            Logger.INSTANCE.warning("Illegal qualifier %s.", qualifier);
            return null;
        }
        Gui gui = this.getGui(split[0]);
        if(gui == null) {
            Logger.INSTANCE.warning("Unknown gui in qualifier %s.", qualifier);
            return null;            
        }
        if(qualifier.charAt(split[0].length()) == '#') {
            return gui.getComponentById(split[1]);
        } else {
            return gui.findByQualifier(split[1]);
        }
    }
    
    /**
     * Liefert eine Gui anhand ihres Namens.
     * @param name Der Name der Gui
     * @return Die Gui
     */
    public Gui getGui(String name) {
        return this.guis.get(name);
    }
    
    public void setDrawer(ComponentDrawer drawer) {
    	this.drawer = drawer;
    }

    /**
     * Fuegt ein ScreenElement hinzu. Die erste hinzugefuegte Szene wird aktiv.
     * @param element Das Element
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean pushElement(ScreenElement element) {
        if(this.screenElements.contains(element)) {
            Logger.INSTANCE.error("Tried to push ScreenElement twice!");
            return false;
        }
        if(!element.onRestore()) {
            Logger.INSTANCE.error("Initializing of ScreenElement failed");
            return false;
        }
        this.screenElements.add(element);
        if(element instanceof SceneScreenElement) {
            SceneScreenElement sceneElement = (SceneScreenElement)element;
            GameApp.getEventManager().register(sceneElement, ActorDeletedEvent.ID);
            if(this.scene == null) {
                this.changeScene(sceneElement.getName());
            }
        }
        return true;
    }
    
    /**
     * Entfernt ein ScreenElement.
     * @param element Das Element
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean removeElement(ScreenElement element) {
        if(this.screenElements.remove(element)) {
            if(element instanceof SceneScreenElement) {
                SceneScreenElement sceneElement = (SceneScreenElement)element;
                GameApp.getEventManager().unregister(sceneElement, ActorDeletedEvent.ID);
                if(this.scene == sceneElement) {
                    GameApp.getEventManager().unregister(sceneElement, ActorCreatedEvent.ID);
                }
            }
            return true;
        } else {
            return false;
        }
    }
    
    @Override
    public void onUpdate(long deltaMillis) {
        boolean inOrder = true;
        int z = Integer.MIN_VALUE;
        for(ScreenElement elem : this.screenElements) {
            if(elem.isVisible()) {
                elem.onUpdate(deltaMillis);
            }
            inOrder &= z <= elem.getzOrder();
            z = elem.getzOrder();
        }
        if(!inOrder) {
            Logger.INSTANCE.info("Rearrange screen elements.");
            Collections.sort(this.screenElements);
        }
    }

    @Override
    public void onRender() {
        for(ScreenElement elem : this.screenElements) {
            if(elem.isVisible()) {
                elem.onRender();
            }
        }
    }

    @Override
    public boolean onLostDevice() {
        boolean success = true;
        for(ScreenElement elem : this.screenElements) {
            if(elem.isVisible()) {
                if(!elem.onLostDevice()) {
                    success = false;
                }
            }
        }
        return success;
    }

    @Override
    public boolean onRestore() {
        boolean success = true;
        for(ScreenElement elem : this.screenElements) {
            if(elem.isVisible()) {
                if(!elem.onRestore()) {
                    success = false;
                }
            }
        }
        return success;
    }

    /**
     * Liefert die aktuelle Szene.
     * @return Die aktuelle Szene
     */
    public SceneScreenElement getScene() {
        return this.scene;
    }

    @Override
    public void trigger(EventData data) {
        if(data instanceof ActorEvent) {
            Actor actor = GameApp.getGameLogic().getActor(((ActorEvent)data).getActorId());
            if(actor != null) {
                if(data.getId() == ActorCreatedEvent.ID) {
                    CameraComponent comp = (CameraComponent)actor.getComponent(CameraComponent.ID);
                    if(comp != null && this.playerActor == null) {
                        this.setPlayer(actor.getName());
                    }
                } else if(data.getId() == ActorDeletedEvent.ID) {
                    if(this.playerActor != null && this.playerActor.getId() == ((ActorEvent)data).getActorId()) {
                        this.setPlayer(null);
                    }
                } 
            }
        }
    }
    
    /**
     * Setzt den Spieleractor.
     * @param actorName Der Name des Actors
     */
    public boolean setPlayer(String actorName) {
        if(this.scene == null) {
            Logger.INSTANCE.warning("Setting player failed, because no scene is active.");
            return false;
        }
        this.playerActor = GameApp.getGameLogic().getActor(actorName);
        GameApp.getEventManager().queueEvent(new PlayerActorChangedEvent(this.playerActor == null ? ActorFactory.INVALID_ACTOR_ID : this.playerActor.getId()));
        if(this.playerActor == null) {
            Logger.INSTANCE.warning("Setting player actor failed, because there is no actor with name " + actorName + ".");
            return false;
        } else {
            CameraComponent comp = (CameraComponent)this.playerActor.getComponent(CameraComponent.ID);
            if(comp == null) {
                Logger.INSTANCE.warning("Set player has no CameraComponent.");
            } else if(this.scene != null) {
                this.scene.setCamera(comp.getPerspectiveCamera());
            }
            return true;
        }
    }
    
    /**
     * Aendert die aktuelle Szene.
     * @param sceneName Der Name der neuen Szene
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean changeScene(String sceneName) {
        for(ScreenElement elem : this.screenElements) {
            if(elem instanceof SceneScreenElement && ((SceneScreenElement)elem).getName().equals(sceneName)) {
                if(this.scene != null) {
                    this.scene.setVisible(false);
                    GameApp.getEventManager().unregister(this.scene, ActorCreatedEvent.ID);
                }
                this.scene = (SceneScreenElement)elem;
                GameApp.getEventManager().register(this.scene, ActorCreatedEvent.ID);
                this.scene.setVisible(true);
                Logger.INSTANCE.info("Scene " + sceneName + " activated.");
                return true;
            }
        }
        Logger.INSTANCE.error("Unknown scene: " + sceneName);
        return false;
    }
}
